import os
import random
import time
from typing import Any, Callable, Optional, Union

import gymnasium as gym
import imageio
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from tqdm import tqdm

import wandb


# ===== CONFIGURATION =====
class Config:
    """Configuration class for A2C training."""

    # Experiment settings
    exp_name: str = "A2C"  # Experiment name for logging and saving
    seed: int = 42  # Random seed for reproducibility
    env_id: Optional[str] = "LunarLander-v3"  # Gymnasium environment ID
    total_timesteps: int = 1000000  # Total timesteps for training (for compatibility)

    # A2C & Agent settings
    lr: float = 2e-3  # Learning rate for optimizer
    gamma: float = 0.99  # Discount factor for rewards
    VALUE_COEFF: float = (
        0.5  # Value loss coefficient in total loss (not used in pure A2C)
    )
    num_eval_episodes: int = 10  # Number of evaluation episodes
    max_grad_norm: float = (
        0.0  # Maximum gradient norm for gradient clipping (0.0 to disable)
    )

    # PPO-specific (kept for compatibility but not used in pure A2C)
    n_envs: int = 1  # Number of parallel environments
    update_epochs: int = 1  # Update epochs (A2C uses 1)
    clip_value: float = 0.2  # Not used in A2C
    ENTROPY_COEFF: float = 0.01  # Not used in pure A2C
    anneal_lr: bool = False  # Learning rate annealing

    # Logging & Saving
    capture_video: bool = True  # Whether to capture evaluation videos
    use_wandb: bool = True  # Whether to use Weights & Biases for logging
    wandb_project: str = "cleanRL"  # W&B project name
    grid_env: bool = False  # Whether the environment uses discrete grid observations (applies OneHot wrapper)
    eval_every: int = 10000  # Frequency of evaluation during training (in updates)
    save_every: int = 10000  # Frequency of saving the model (in updates)
    atari_wrapper: bool = (
        False  # Whether to apply Atari preprocessing and frame stacking
    )
    normalize_obs: bool = False  # Whether to normalize observations
    normalize_reward: bool = False  # Whether to normalize rewards
    log_gradients: bool = False  # Whether to log gradient norms to W&B
    device: str = "cpu"  # Device for training: "auto", "cpu", "cuda", or "cuda:0" etc.
    custom_agent: Optional[nn.Module] = None  # Custom neural network class or instance
    env: Optional[gym.Env] = None  # Optional pre-created environment


# ===== RunningMeanStd for Normalization =====
class RunningMeanStd:
    """Tracks running mean and std of a data stream using Welford's algorithm."""

    def __init__(self, epsilon: float = 1e-4, shape: tuple[int, ...] = ()):
        self.mean = np.zeros(shape, dtype=np.float64)
        self.var = np.ones(shape, dtype=np.float64)
        self.count = epsilon

    def update(self, x: np.ndarray) -> None:
        """Update running statistics with new batch of data."""
        batch_mean = np.mean(x, axis=0)
        batch_var = np.var(x, axis=0)
        batch_count = x.shape[0]
        self.update_from_moments(batch_mean, batch_var, batch_count)

    def update_from_moments(
        self, batch_mean: np.ndarray, batch_var: np.ndarray, batch_count: int
    ) -> None:
        """Update from precomputed moments."""
        delta = batch_mean - self.mean
        tot_count = self.count + batch_count

        new_mean = self.mean + delta * batch_count / tot_count
        m_a = self.var * self.count
        m_b = batch_var * batch_count
        M2 = m_a + m_b + np.square(delta) * self.count * batch_count / tot_count
        new_var = M2 / tot_count

        self.mean = new_mean
        self.var = new_var
        self.count = tot_count


# --- Networks ---
def layer_init(
    layer: nn.Module, std: float = np.sqrt(2), bias_const: float = 0.0
) -> nn.Module:
    torch.nn.init.orthogonal_(layer.weight, std)
    torch.nn.init.constant_(layer.bias, bias_const)
    return layer


class ActorNet(nn.Module):
    def __init__(
        self,
        state_space: Union[int, tuple[int, ...]],
        action_space: int,
    ) -> None:
        super().__init__()

        self.fc1 = layer_init(nn.Linear(state_space, 32))
        self.fc2 = layer_init(nn.Linear(32, 32))
        self.fc3 = layer_init(nn.Linear(32, 16))
        self.out = layer_init(nn.Linear(16, action_space))

    def forward(self, x: torch.Tensor) -> torch.distributions.Distribution:
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))

        logits = self.out(x)
        probs = torch.nn.functional.softmax(logits, dim=-1)
        dist = torch.distributions.Categorical(probs)
        return dist

    def get_action(
        self,
        x: torch.Tensor,
        action: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.distributions.Distribution]:
        dist = self.forward(x)
        if action is None:
            action = dist.sample()

        log_prob = dist.log_prob(action)

        return action, log_prob, dist


class CriticNet(nn.Module):
    def __init__(self, state_space: Union[int, tuple[int, ...]]) -> None:
        super().__init__()
        self.fc1 = layer_init(nn.Linear(state_space, 32))
        self.fc2 = layer_init(nn.Linear(32, 32))
        self.fc3 = layer_init(nn.Linear(32, 16))
        self.value = layer_init(nn.Linear(16, 1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))
        return self.value(x)


class OneHotWrapper(gym.ObservationWrapper):
    def __init__(self, env: gym.Env, obs_shape: int = 16) -> None:
        super().__init__(env)
        self.obs_shape = obs_shape
        self.observation_space = gym.spaces.Box(0, 1, (obs_shape,), dtype=np.float32)

    def observation(self, obs: Any) -> np.ndarray:
        one_hot = torch.zeros(self.obs_shape, dtype=torch.float32)
        one_hot[obs] = 1.0
        return one_hot.numpy()


def make_env(
    env_id: str,
    seed: int,
    idx: int,
    render_mode: Optional[str] = None,
    grid_env: bool = False,
    atari_wrapper: bool = False,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    env: Optional[gym.Env] = None,
) -> Callable[[], gym.Env]:
    def thunk():
        if env is not None:
            # Use provided environment
            env_to_use = env

        else:
            # Create new environment
            env_to_use = gym.make(env_id, render_mode=render_mode)

        env_to_use = gym.wrappers.RecordEpisodeStatistics(env_to_use)
        if Config.normalize_reward:
            env_to_use = gym.wrappers.NormalizeReward(env_to_use)
        if grid_env:
            env_to_use = OneHotWrapper(
                env_to_use, obs_shape=env_to_use.observation_space.n
            )
        if Config.normalize_obs:
            env_to_use = gym.wrappers.NormalizeObservation(env_to_use)
        if atari_wrapper:
            env_to_use = gym.wrappers.AtariPreprocessing(
                env_to_use, grayscale_obs=True, scale_obs=True
            )
            env_to_use = gym.wrappers.FrameStackObservation(env_to_use, stack_size=4)
        if env_wrapper:
            env_to_use = env_wrapper(env_to_use)
        env_to_use.action_space.seed(seed + idx)
        return env_to_use

    return thunk


def calculate_param_norm(model: nn.Module) -> float:
    """Calculate the L2 norm of all parameters in a model."""
    total_norm = 0.0
    for p in model.parameters():
        param_norm = p.data.norm(2)
        total_norm += param_norm.item() ** 2
    return total_norm**0.5


def validate_policy_network_dimensions(
    policy_network: nn.Module, obs_dim: Union[int, tuple[int, ...]], action_dim: int
) -> None:
    """
    Validate that the Policy-network's input and output dimensions match the environment.

    Args:
        policy_network: The neural network model (nn.Module)
        obs_dim: Expected observation dimension (int or tuple)
        action_dim: Expected action dimension
    """
    if isinstance(obs_dim, tuple):
        # For Atari-like, check if it has conv layers
        has_conv = any(
            isinstance(module, nn.Conv2d) for module in policy_network.modules()
        )
        if not has_conv:
            print(
                "Warning: Observation is multi-dimensional but network has no Conv2d layers."
            )
    else:
        # Find first Linear layer for input dimension
        first_layer = None
        for module in policy_network.modules():
            if isinstance(module, nn.Linear):
                first_layer = module
                break
        if first_layer is None:
            raise ValueError(
                "Policy-network must have at least one Linear layer for dimension validation."
            )
        if first_layer.in_features != obs_dim:
            raise ValueError(
                f"Policy-network input dimension {first_layer.in_features} does not match observation dimension {obs_dim}."
            )

    # Find last Linear layer for output dimension
    last_layer = None
    for module in reversed(list(policy_network.modules())):
        if isinstance(module, nn.Linear):
            last_layer = module
            break
    if last_layer is None:
        raise ValueError(
            "Policy-network must have at least one Linear layer for dimension validation."
        )
    if last_layer.out_features != action_dim:
        raise ValueError(
            f"Policy-network output dimension {last_layer.out_features} does not match action dimension {action_dim}."
        )


def validate_critic_network_dimensions(
    critic_network: nn.Module, obs_dim: Union[int, tuple[int, ...]]
) -> None:
    """
    Validate that the Critic-network's input dimension matches the environment.

    Args:
        critic_network: The critic neural network model (nn.Module)
        obs_dim: Expected observation dimension (int or tuple)
    """
    if isinstance(obs_dim, tuple):
        # For Atari-like, check if it has conv layers
        has_conv = any(
            isinstance(module, nn.Conv2d) for module in critic_network.modules()
        )
        if not has_conv:
            print(
                "Warning: Observation is multi-dimensional but network has no Conv2d layers."
            )
    else:
        # Find first Linear layer for input dimension
        first_layer = None
        for module in critic_network.modules():
            if isinstance(module, nn.Linear):
                first_layer = module
                break
        if first_layer is None:
            raise ValueError(
                "Critic-network must have at least one Linear layer for dimension validation."
            )
        if first_layer.in_features != obs_dim:
            raise ValueError(
                f"Critic-network input dimension {first_layer.in_features} does not match observation dimension {obs_dim}."
            )

    # Find last Linear layer for output dimension (should be 1 for value)
    last_layer = None
    for module in reversed(list(critic_network.modules())):
        if isinstance(module, nn.Linear):
            last_layer = module
            break
    if last_layer is None:
        raise ValueError(
            "Critic-network must have at least one Linear layer for dimension validation."
        )
    if last_layer.out_features != 1:
        raise ValueError(
            f"Critic-network output dimension {last_layer.out_features} should be 1 for value estimation."
        )


def validate_feature_network_dimensions(
    feature_network: nn.Module, obs_dim: Union[int, tuple[int, ...]], feature_dim: int
) -> None:
    """
    Validate that the Feature-network's input and output dimensions match expectations.

    Args:
        feature_network: The feature neural network model (nn.Module)
        obs_dim: Expected observation dimension (int or tuple)
        feature_dim: Expected feature dimension
    """
    if isinstance(obs_dim, tuple):
        # For Atari-like, check if it has conv layers
        has_conv = any(
            isinstance(module, nn.Conv2d) for module in feature_network.modules()
        )
        if not has_conv:
            print(
                "Warning: Observation is multi-dimensional but network has no Conv2d layers."
            )
    else:
        # Find first Linear layer for input dimension
        first_layer = None
        for module in feature_network.modules():
            if isinstance(module, nn.Linear):
                first_layer = module
                break
        if first_layer is None:
            raise ValueError(
                "Feature-network must have at least one Linear layer for dimension validation."
            )
        if first_layer.in_features != obs_dim:
            raise ValueError(
                f"Feature-network input dimension {first_layer.in_features} does not match observation dimension {obs_dim}."
            )

    # Find last Linear layer for output dimension
    last_layer = None
    for module in reversed(list(feature_network.modules())):
        if isinstance(module, nn.Linear):
            last_layer = module
            break
    if last_layer is None:
        raise ValueError(
            "Feature-network must have at least one Linear layer for dimension validation."
        )
    if last_layer.out_features != feature_dim:
        raise ValueError(
            f"Feature-network output dimension {last_layer.out_features} does not match expected feature dimension {feature_dim}."
        )


def evaluate(
    model: nn.Module,
    device: torch.device,
    env_id: str,
    env: Optional[gym.Env] = None,
    seed: int = 42,
    num_eval_eps: int = 5,
    record: bool = False,
    render_mode: Optional[str] = None,
    grid_env: bool = False,
    atari_wrapper: bool = False,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    use_wandb: bool = False,
) -> tuple[list[float], list[np.ndarray]]:
    # Create evaluation environment
    eval_env = make_env(
        env_id if env is None else "",
        seed,
        idx=0,
        render_mode=render_mode,
        grid_env=grid_env,
        atari_wrapper=atari_wrapper,
        env_wrapper=env_wrapper,
        env=env,
    )()
    model.eval()
    returns = []
    frames = []

    for _ in tqdm(range(num_eval_eps), desc="Evaluating"):
        obs, _ = eval_env.reset()
        done = False
        rewards = 0.0

        while not done:
            if record:
                frame = eval_env.render()
                frames.append(frame)
            with torch.no_grad():
                obs = torch.tensor(obs, device=device, dtype=torch.float32)
                action, _, _ = model.get_action(obs)
                action = action.cpu().numpy()
                if isinstance(eval_env.action_space, gym.spaces.Discrete):
                    action = action.item()

            obs, rewards_curr, terminated, truncated, info = eval_env.step(action)
            done = terminated or truncated
            rewards += rewards_curr

        returns.append(rewards)

    # Save video
    if frames and use_wandb:
        video = np.stack(frames)

        video = np.transpose(video, (0, 3, 1, 2))

        wandb.log(
            {
                "videos/eval_policy": wandb.Video(
                    video,
                    fps=30,
                    format="mp4",
                )
            }
        )
        frames = []
    eval_env.close()
    model.train()

    return returns, frames


def train_a2c(
    env_id: Optional[str] = None,
    env: Optional[gym.Env] = Config.env,
    total_timesteps: int = Config.total_timesteps,
    seed: int = Config.seed,
    lr: float = Config.lr,
    gamma: float = Config.gamma,
    n_envs: int = Config.n_envs,
    ENTROPY_COEFF: float = Config.ENTROPY_COEFF,
    capture_video: bool = Config.capture_video,
    use_wandb: bool = Config.use_wandb,
    wandb_project: str = Config.wandb_project,
    exp_name: str = Config.exp_name,
    grid_env: bool = Config.grid_env,
    max_grad_norm: float = Config.max_grad_norm,
    eval_every: int = Config.eval_every,
    save_every: int = Config.save_every,
    atari_wrapper: bool = Config.atari_wrapper,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    actor_class: Any = ActorNet,
    critic_class: Any = CriticNet,
    log_gradients: bool = Config.log_gradients,
    num_eval_episodes: int = Config.num_eval_episodes,
    anneal_lr: bool = Config.anneal_lr,
    normalize_obs: bool = Config.normalize_obs,
    normalize_reward: bool = Config.normalize_reward,
    device: str = Config.device,
) -> nn.Module:
    # Update Config with passed arguments
    Config.env_id = env_id if env is None else env.spec.id
    Config.total_timesteps = total_timesteps
    Config.seed = seed
    Config.lr = lr
    Config.gamma = gamma
    Config.n_envs = n_envs
    Config.ENTROPY_COEFF = ENTROPY_COEFF
    Config.capture_video = capture_video
    Config.use_wandb = use_wandb
    Config.wandb_project = wandb_project
    Config.exp_name = exp_name
    Config.grid_env = grid_env
    Config.max_grad_norm = max_grad_norm
    Config.eval_every = eval_every
    Config.save_every = save_every
    Config.atari_wrapper = atari_wrapper
    Config.log_gradients = log_gradients
    Config.num_eval_episodes = num_eval_episodes
    Config.anneal_lr = anneal_lr
    Config.normalize_obs = normalize_obs
    Config.normalize_reward = normalize_reward
    Config.device = device

    if env is not None and Config.env_id != env_id:
        raise ValueError(
            "Cannot provide both env_id and env. Use env_id for default environments or env for custom environments."
        )

    if Config.capture_video and not Config.use_wandb:
        raise ValueError(
            "Cannot capture video without WandB enabled. Set use_wandb=True to upload videos."
        )

    run_name = f"{Config.env_id}__{Config.exp_name}__{Config.seed}__{int(time.time())}"

    os.makedirs(f"runs/{run_name}/models", exist_ok=True)

    if Config.use_wandb:
        wandb.init(
            project=Config.wandb_project,
            config=vars(Config()),
            name=run_name,
            monitor_gym=True,
        )

    random.seed(Config.seed)
    np.random.seed(Config.seed)
    torch.manual_seed(Config.seed)

    device = torch.device(Config.device)

    if Config.device == "cuda":
        torch.backends.cudnn.deterministic = True
        torch.cuda.manual_seed(Config.seed)
        torch.cuda.manual_seed_all(Config.seed)
        torch.backends.cudnn.benchmark = False

    # Create environments - check for pre-created env first, then default
    if env is not None:
        env_thunks = [
            make_env(
                "",
                Config.seed,
                i,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                env=env,
                render_mode="rgb_array",
            )
            for i in range(Config.n_envs)
        ]
    else:
        # Use default environment creation
        env_thunks = [
            make_env(
                Config.env_id,
                Config.seed,
                i,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                render_mode="rgb_array",
            )
            for i in range(Config.n_envs)
        ]

    envs = gym.vector.SyncVectorEnv(env_thunks)
    if isinstance(envs.single_observation_space, gym.spaces.Discrete):
        obs_space_shape = (envs.single_observation_space.n,)
    else:
        obs_space_shape = envs.single_observation_space.shape[0]

    action_space_n = (
        envs.single_action_space.n
        if isinstance(envs.single_action_space, gym.spaces.Discrete)
        else envs.single_action_space.shape[0]
    )

    print(f"Observation Space: {obs_space_shape}, Action Space: {action_space_n}")

    # Create actor network
    if isinstance(actor_class, nn.Module):
        # Use custom actor instance
        validate_policy_network_dimensions(actor_class, obs_space_shape, action_space_n)
        actor_network = actor_class.to(device)
    else:
        # Use actor class
        actor_network = actor_class(obs_space_shape, action_space_n).to(device)

    # Create critic network
    if isinstance(critic_class, nn.Module):
        # Use custom critic instance
        validate_critic_network_dimensions(critic_class, obs_space_shape)
        critic_network = critic_class.to(device)
    else:
        # Use critic class
        critic_network = critic_class(obs_space_shape).to(device)

    print("Actor Network Architecture:")
    print(actor_network)
    print("\nCritic Network Architecture:")
    print(critic_network)

    num_updates = Config.total_timesteps // Config.n_envs

    # Separate optimizers for actor and critic (A2C style)
    actor_optim = optim.Adam(actor_network.parameters(), lr=Config.lr)
    critic_optim = optim.Adam(critic_network.parameters(), lr=Config.lr)

    global_step = 0

    next_obs, _ = envs.reset(seed=Config.seed)
    next_obs = torch.Tensor(next_obs).to(device)

    start_time = time.time()

    for step in tqdm(range(1, num_updates + 1), desc="Training Updates"):
        # Annealing the rate if instructed to do so.
        global_step = step * Config.n_envs
        if Config.anneal_lr:
            frac = 1.0 - (step - 1.0) / num_updates
            lrnow = frac * Config.lr
            actor_optim.param_groups[0]["lr"] = lrnow
            critic_optim.param_groups[0]["lr"] = lrnow

        # Rollout Phase - Collect full episode

        log_probs = []
        rewards = []
        values = []
        entropies = []

        while True:
            # Get action and value
            result = actor_network.get_action(next_obs)
            if len(result) == 2:
                action, logprob = result
                logprob = logprob.sum(dim=-1) if len(logprob.shape) > 1 else logprob
                dist = None
            elif len(result) == 3:
                action, logprob, dist = result
                logprob = logprob.sum(dim=-1) if len(logprob.shape) > 1 else logprob
            else:
                raise ValueError(
                    f"Error unpacking result from get_action. Expected 2 or 3 values, got {len(result)}"
                )

            # Track entropy if enabled
            if Config.ENTROPY_COEFF > 0.0 and dist is not None:
                entropies.append(dist.entropy())

            value = critic_network(next_obs)

            # Log distribution statistics
            if Config.use_wandb and step % 100 == 0:
                if dist is not None and hasattr(dist, "probs"):
                    # Discrete
                    probs = dist.probs
                    wandb.log(
                        {
                            "policy/dist_mean": probs.mean().item(),
                            "policy/dist_std": probs.std().item(),
                            "policy/entropy": dist.entropy().mean().item(),
                            "global_step": global_step,
                        }
                    )

            values.append(value.flatten())

            log_probs.append(logprob)

            # Step the environment
            new_obs, reward, terminated, truncated, info = envs.step(
                action.cpu().numpy()
            )
            done = np.logical_or(terminated, truncated)

            rewards.append(
                torch.tensor(reward, dtype=torch.float32).to(device).view(-1)
            )
            next_obs = torch.Tensor(new_obs).to(device)

            if np.all(done):
                break

        # Convert lists to tensors

        log_probs_tensor = torch.stack(log_probs)
        rewards_tensor = torch.stack(rewards)

        values_tensor = torch.stack(values)

        # Calculate returns (Monte Carlo)
        num_steps = len(rewards)
        returns = torch.zeros_like(rewards_tensor, dtype=torch.float32).to(device)
        rt = 0.0
        for t in reversed(range(num_steps)):
            rt = rewards_tensor[t] + rt * Config.gamma
            returns[t] = rt

        # Calculate advantages (no normalization in pure A2C)
        advantages = (returns - values_tensor).detach()

        # Flatten for batch processing
        b_logprobs = log_probs_tensor.flatten()
        b_values = values_tensor.flatten()
        b_returns = returns.flatten()
        b_advantages = advantages.flatten()

        # A2C Update - Single pass through data
        # Actor Loss: policy gradient with advantage
        policy_loss = -(b_logprobs * b_advantages).mean()

        # Add entropy bonus if enabled
        if Config.ENTROPY_COEFF > 0.0 and entropies:
            entropy_loss = torch.stack(entropies).mean() * Config.ENTROPY_COEFF
            policy_loss = policy_loss - entropy_loss
            if Config.use_wandb:
                wandb.log(
                    {
                        "losses/entropy_loss": entropy_loss.item(),
                        "global_step": global_step,
                    }
                )

        # Critic Loss: MSE between value predictions and returns
        critic_loss = F.mse_loss(b_values, b_returns)

        # Optimize Actor
        actor_optim.zero_grad()
        policy_loss.backward()

        # Log gradient norms for actor
        if Config.use_wandb and Config.log_gradients:
            total_norm = 0
            for name, param in actor_network.named_parameters():
                if param.grad is not None:
                    param_norm = param.grad.data.norm(2)
                    wandb.log(
                        {
                            f"gradients/actor_{name}": param_norm.item(),
                            "global_step": global_step,
                        }
                    )
                    total_norm += param_norm.item() ** 2
            wandb.log(
                {
                    "gradients/actor_total_norm": total_norm**0.5,
                    "global_step": global_step,
                }
            )

        if Config.max_grad_norm != 0.0:
            torch.nn.utils.clip_grad_norm_(
                actor_network.parameters(), max_norm=Config.max_grad_norm
            )
        actor_optim.step()

        # Optimize Critic
        critic_optim.zero_grad()
        critic_loss.backward()

        # Log gradient norms for critic
        if Config.use_wandb and Config.log_gradients:
            total_norm = 0
            for name, param in critic_network.named_parameters():
                if param.grad is not None:
                    param_norm = param.grad.data.norm(2)
                    wandb.log(
                        {
                            f"gradients/critic_{name}": param_norm.item(),
                            "global_step": global_step,
                        }
                    )
                    total_norm += param_norm.item() ** 2
            wandb.log(
                {
                    "gradients/critic_total_norm": total_norm**0.5,
                    "global_step": global_step,
                }
            )

        if Config.max_grad_norm != 0.0:
            torch.nn.utils.clip_grad_norm_(
                critic_network.parameters(), max_norm=Config.max_grad_norm
            )
        critic_optim.step()

        # Log episode returns
        if "episode" in info:
            if Config.n_envs > 1:
                for i in range(Config.n_envs):
                    if done[i]:
                        ep_ret = info["episode"]["r"][i]
                        ep_len = info["episode"]["l"][i]

                        if Config.use_wandb:
                            wandb.log(
                                {
                                    "charts/episodic_return": ep_ret,
                                    "charts/episodic_length": ep_len,
                                    "global_step": global_step,
                                }
                            )
            else:
                if done:
                    ep_ret = info["episode"]["r"]
                    ep_len = info["episode"]["l"]
                    if Config.use_wandb:
                        wandb.log(
                            {
                                "charts/episodic_return": ep_ret,
                                "charts/episodic_length": ep_len,
                                "global_step": global_step,
                            }
                        )

        # Log losses and metrics
        if Config.use_wandb and step % 10 == 0:
            wandb.log(
                {
                    "losses/policy_loss": policy_loss.item(),
                    "losses/critic_loss": critic_loss.item(),
                    "charts/learning_rate": actor_optim.param_groups[0]["lr"],
                    "charts/rewards_mean": rewards_tensor.mean().item(),
                    "advantages/advantages_mean": advantages.mean().item(),
                    "advantages/advantages_std": advantages.std().item(),
                    "charts/returns_mean": b_returns.mean().item(),
                    "global_step": global_step,
                }
            )
            print(
                f"Step {step}, Global Step: {global_step}, Policy Loss: {policy_loss.item():.4f}, Critic Loss: {critic_loss.item():.4f}, SPS: {int(global_step / (time.time() - start_time))}"
            )

            if Config.use_wandb:
                wandb.log(
                    {
                        "charts/SPS": int(global_step / (time.time() - start_time)),
                        "charts/global_step": global_step,
                    }
                )
        if step % Config.eval_every == 0:
            episodic_returns, _ = evaluate(
                actor_network,
                device,
                Config.env_id,
                env=env,
                seed=Config.seed,
                num_eval_eps=Config.num_eval_episodes,
                record=Config.capture_video,
                render_mode="rgb_array" if Config.capture_video else None,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                use_wandb=Config.use_wandb,
            )
            avg_return = np.mean(episodic_returns)
            if Config.use_wandb:
                wandb.log(
                    {"charts/val_avg_return": avg_return, "global_step": global_step}
                )
            print(f"Evaluation returns: {episodic_returns}, Average: {avg_return:.2f}")

        if step % Config.save_every == 0 and step > 0:
            model_path = f"runs/{run_name}/models/a2c_model_step_{global_step}.pth"
            torch.save(
                {
                    "actor": actor_network.state_dict(),
                    "critic": critic_network.state_dict(),
                },
                model_path,
            )
            print(f"Model saved at step {step} step {global_step} to {model_path}")

    # --- Final Evaluation and Video Saving ---
    if Config.use_wandb:
        print("Capturing final evaluation video...")
        episodic_returns, eval_frames = evaluate(
            actor_network,
            device,
            Config.env_id,
            env=env,
            seed=Config.seed,
            num_eval_eps=Config.num_eval_episodes,
            record=Config.capture_video,
            grid_env=Config.grid_env,
            atari_wrapper=Config.atari_wrapper,
            env_wrapper=env_wrapper,
            use_wandb=Config.use_wandb,
            render_mode="rgb_array" if Config.capture_video else None,
        )

        # Save final video to file if frames were captured
        if eval_frames:
            train_video_path = "videos/final.mp4"
            imageio.mimsave(train_video_path, eval_frames, fps=30)
            print(f"Final training video saved to {train_video_path}")

    envs.close()
    if Config.use_wandb:
        wandb.finish()

    return actor_network


def train_a2c_cnn(
    env_id: str = Config.env_id,
    env: Optional[gym.Env] = None,
    total_timesteps: int = Config.total_timesteps,
    seed: int = Config.seed,
    lr: float = Config.lr,
    gamma: float = Config.gamma,
    n_envs: int = Config.n_envs,
    ENTROPY_COEFF: float = Config.ENTROPY_COEFF,
    VALUE_COEFF: float = Config.VALUE_COEFF,
    capture_video: bool = Config.capture_video,
    use_wandb: bool = Config.use_wandb,
    wandb_project: str = Config.wandb_project,
    exp_name: str = Config.exp_name,
    grid_env: bool = Config.grid_env,
    max_grad_norm: float = Config.max_grad_norm,
    eval_every: int = Config.eval_every,
    save_every: int = Config.save_every,
    atari_wrapper: bool = Config.atari_wrapper,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    actor_class: Any = ActorNet,
    critic_class: Any = CriticNet,
    log_gradients: bool = Config.log_gradients,
    num_eval_episodes: int = Config.num_eval_episodes,
    anneal_lr: bool = Config.anneal_lr,
    normalize_obs: bool = Config.normalize_obs,
    normalize_reward: bool = Config.normalize_reward,
    device: str = Config.device,
) -> nn.Module:
    # Update Config with passed arguments
    Config.env_id = env_id if env is None else env.spec.id
    Config.total_timesteps = total_timesteps
    Config.seed = seed
    Config.lr = lr
    Config.gamma = gamma
    Config.n_envs = n_envs
    Config.ENTROPY_COEFF = ENTROPY_COEFF
    Config.VALUE_COEFF = VALUE_COEFF
    Config.capture_video = capture_video
    Config.use_wandb = use_wandb
    Config.wandb_project = wandb_project
    Config.exp_name = exp_name
    Config.grid_env = grid_env
    Config.max_grad_norm = max_grad_norm
    Config.eval_every = eval_every
    Config.save_every = save_every
    Config.atari_wrapper = atari_wrapper
    Config.log_gradients = log_gradients
    Config.num_eval_episodes = num_eval_episodes
    Config.anneal_lr = anneal_lr
    Config.normalize_obs = normalize_obs
    Config.normalize_reward = normalize_reward
    Config.device = device

    if env is not None and Config.env_id != env_id:
        raise ValueError(
            "Cannot provide both env_id and env. Use env_id for default environments or env for custom environments."
        )

    if Config.capture_video and not Config.use_wandb:
        raise ValueError(
            "Cannot capture video without WandB enabled. Set use_wandb=True to upload videos."
        )

    run_name = f"{Config.env_id}__{Config.exp_name}__{Config.seed}__{int(time.time())}"

    os.makedirs(f"runs/{run_name}/models", exist_ok=True)

    if Config.use_wandb:
        wandb.init(
            project=Config.wandb_project,
            config=vars(Config()),
            name=run_name,
            monitor_gym=True,
        )

    random.seed(Config.seed)
    np.random.seed(Config.seed)
    torch.manual_seed(Config.seed)
    if Config.device == "auto":
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    else:
        device = torch.device(Config.device)

    # Create environments - check for pre-created env first, then default
    if env is not None:
        env_thunks = [
            make_env(
                "",
                Config.seed,
                i,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                env=env,
                render_mode="rgb_array",
            )
            for i in range(Config.n_envs)
        ]
    else:
        # Use default environment creation
        env_thunks = [
            make_env(
                Config.env_id,
                Config.seed,
                i,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                render_mode="rgb_array",
            )
            for i in range(Config.n_envs)
        ]

    envs = gym.vector.SyncVectorEnv(env_thunks)
    if isinstance(envs.single_observation_space, gym.spaces.Discrete):
        obs_space_shape = (envs.single_observation_space.n,)
    else:
        obs_space_shape = envs.single_observation_space.shape

    action_space_n = (
        envs.single_action_space.n
        if isinstance(envs.single_action_space, gym.spaces.Discrete)
        else envs.single_action_space.shape
    )

    print(f"Observation Space: {obs_space_shape}, Action Space: {action_space_n}")

    # Create actor network
    if isinstance(actor_class, nn.Module):
        # Use custom actor instance
        validate_policy_network_dimensions(
            actor_class,
            obs_space_shape[0] if len(obs_space_shape) == 1 else obs_space_shape,
            action_space_n,
        )
        actor_network = actor_class.to(device)
    else:
        # Use actor class
        actor_network = actor_class(obs_space_shape, action_space_n).to(device)

    # Create critic network
    if isinstance(critic_class, nn.Module):
        # Use custom critic instance
        validate_critic_network_dimensions(
            critic_class,
            obs_space_shape[0] if len(obs_space_shape) == 1 else obs_space_shape,
        )
        critic_network = critic_class.to(device)
    else:
        # Use critic class
        critic_network = critic_class(obs_space_shape).to(device)

    print("Actor Network Architecture:")
    print(actor_network)
    print("\nCritic Network Architecture:")
    print(critic_network)

    # Compute derived values from passed parameters
    batch_size = Config.n_envs * Config.max_steps
    num_updates = Config.total_timesteps // batch_size

    # Separate optimizers for actor and critic (A2C style)
    actor_optim = optim.Adam(actor_network.parameters(), lr=Config.lr)
    critic_optim = optim.Adam(critic_network.parameters(), lr=Config.lr)

    global_step = 0

    next_obs, _ = envs.reset(seed=Config.seed)
    next_obs = torch.Tensor(next_obs).to(device)
    next_done = torch.zeros(Config.n_envs).to(device)

    start_time = time.time()

    for update in tqdm(range(1, num_updates + 1), desc="Training Updates"):
        # Annealing the rate if instructed to do so.
        if Config.anneal_lr:
            frac = 1.0 - (update - 1.0) / num_updates
            lrnow = frac * Config.lr
            actor_optim.param_groups[0]["lr"] = lrnow
            critic_optim.param_groups[0]["lr"] = lrnow

        # Rollout Phase - Collect full episode

        log_probs = []
        rewards = []
        dones = []
        values = []
        entropies = []

        for step in range(0, Config.max_steps):
            global_step = (update - 1) * batch_size + step * Config.n_envs

            dones.append(next_done)

            # Get action and value
            result = actor_network.get_action(next_obs)
            if len(result) == 2:
                action, logprob = result
                logprob = logprob.sum(dim=-1) if len(logprob.shape) > 1 else logprob
                dist = None
            elif len(result) == 3:
                action, logprob, dist = result
                logprob = logprob.sum(dim=-1) if len(logprob.shape) > 1 else logprob
            else:
                raise ValueError(
                    f"Error unpacking result from get_action. Expected 2 or 3 values, got {len(result)}"
                )

            # Track entropy if enabled
            if Config.ENTROPY_COEFF > 0.0 and dist is not None:
                entropies.append(dist.entropy())

            value = critic_network(next_obs)

            # Log distribution statistics
            if Config.use_wandb and step % 100 == 0:
                if dist is not None and hasattr(dist, "probs"):
                    # Discrete
                    probs = dist.probs
                    wandb.log(
                        {
                            "policy/dist_mean": probs.mean().item(),
                            "policy/dist_std": probs.std().item(),
                            "policy/entropy": dist.entropy().mean().item(),
                            "global_step": global_step,
                        }
                    )

            values.append(value.flatten())

            log_probs.append(logprob)

            # Step the environment
            new_obs, reward, terminated, truncated, info = envs.step(
                action.cpu().numpy()
            )
            done = np.logical_or(terminated, truncated)

            rewards.append(
                torch.tensor(reward, dtype=torch.float32).to(device).view(-1)
            )
            next_obs = torch.Tensor(new_obs).to(device)
            next_done = torch.Tensor(done).to(device)

        # Convert lists to tensors

        log_probs_tensor = torch.stack(log_probs)
        rewards_tensor = torch.stack(rewards)
        values_tensor = torch.stack(values)

        # Calculate returns (Monte Carlo)
        num_steps = len(rewards)
        returns = torch.zeros_like(rewards_tensor, dtype=torch.float32).to(device)
        rt = 0.0
        for t in reversed(range(num_steps)):
            rt = rewards_tensor[t] + rt * Config.gamma
            returns[t] = rt

        # Flatten tensors for batch processing
        b_logprobs = log_probs_tensor.reshape(-1)
        b_values = values_tensor.reshape(-1)
        b_returns = returns.reshape(-1)

        # Calculate advantages (no normalization in pure A2C)
        advantages = (b_returns - b_values).detach()

        # A2C Update - Single pass through data
        # Actor Loss: policy gradient with advantage
        policy_loss = -(b_logprobs * advantages).mean()

        # Add entropy bonus if enabled
        if Config.ENTROPY_COEFF > 0.0 and entropies:
            entropy_loss = torch.stack(entropies).mean() * Config.ENTROPY_COEFF
            policy_loss = policy_loss - entropy_loss
            if Config.use_wandb:
                wandb.log(
                    {
                        "losses/entropy_loss": entropy_loss.item(),
                        "global_step": global_step,
                    }
                )

        # Critic Loss: MSE between value predictions and returns
        critic_loss = F.mse_loss(b_values, b_returns)

        # Optimize Actor
        actor_optim.zero_grad()
        policy_loss.backward()

        # Log gradient norms for actor
        if Config.use_wandb and Config.log_gradients:
            total_norm = 0
            for name, param in actor_network.named_parameters():
                if param.grad is not None:
                    param_norm = param.grad.data.norm(2)
                    wandb.log(
                        {
                            f"gradients/actor_{name}": param_norm.item(),
                            "global_step": global_step,
                        }
                    )
                    total_norm += param_norm.item() ** 2
            wandb.log(
                {
                    "gradients/actor_total_norm": total_norm**0.5,
                    "global_step": global_step,
                }
            )

        if Config.max_grad_norm != 0.0:
            torch.nn.utils.clip_grad_norm_(
                actor_network.parameters(), max_norm=Config.max_grad_norm
            )
        actor_optim.step()

        # Optimize Critic
        critic_optim.zero_grad()
        critic_loss.backward()

        # Log gradient norms for critic
        if Config.use_wandb and Config.log_gradients:
            total_norm = 0
            for name, param in critic_network.named_parameters():
                if param.grad is not None:
                    param_norm = param.grad.data.norm(2)
                    wandb.log(
                        {
                            f"gradients/critic_{name}": param_norm.item(),
                            "global_step": global_step,
                        }
                    )
                    total_norm += param_norm.item() ** 2
            wandb.log(
                {
                    "gradients/critic_total_norm": total_norm**0.5,
                    "global_step": global_step,
                }
            )

        if Config.max_grad_norm != 0.0:
            torch.nn.utils.clip_grad_norm_(
                critic_network.parameters(), max_norm=Config.max_grad_norm
            )
        critic_optim.step()

        # Log episode returns
        if "episode" in info:
            if Config.n_envs > 1:
                for i in range(Config.n_envs):
                    if done[i]:
                        ep_ret = info["episode"]["r"][i]
                        ep_len = info["episode"]["l"][i]

                        if Config.use_wandb:
                            wandb.log(
                                {
                                    "charts/episodic_return": ep_ret,
                                    "charts/episodic_length": ep_len,
                                    "global_step": global_step,
                                }
                            )
            else:
                if done:
                    ep_ret = info["episode"]["r"]
                    ep_len = info["episode"]["l"]
                    if Config.use_wandb:
                        wandb.log(
                            {
                                "charts/episodic_return": ep_ret,
                                "charts/episodic_length": ep_len,
                                "global_step": global_step,
                            }
                        )

        # Log losses and metrics
        if Config.use_wandb and update % 10 == 0:
            wandb.log(
                {
                    "losses/policy_loss": policy_loss.item(),
                    "losses/critic_loss": critic_loss.item(),
                    "charts/learning_rate": actor_optim.param_groups[0]["lr"],
                    "charts/rewards_mean": rewards_tensor.mean().item(),
                    "advantages/advantages_mean": advantages.mean().item(),
                    "advantages/advantages_std": advantages.std().item(),
                    "charts/returns_mean": b_returns.mean().item(),
                    "calculated_return": b_returns.mean().item(),
                    "global_step": global_step,
                }
            )
            print(
                f"Update {update}, Global Step: {global_step}, Policy Loss: {policy_loss.item():.4f}, Critic Loss: {critic_loss.item():.4f}, SPS: {int(global_step / (time.time() - start_time))}"
            )

            if Config.use_wandb:
                wandb.log(
                    {
                        "charts/SPS": int(global_step / (time.time() - start_time)),
                        "charts/global_step": global_step,
                    }
                )
        if update % Config.eval_every == 0:
            episodic_returns, _ = evaluate(
                actor_network,
                device,
                Config.env_id,
                env=env,
                seed=Config.seed,
                num_eval_eps=Config.num_eval_episodes,
                record=Config.capture_video,
                render_mode="rgb_array" if Config.capture_video else None,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                use_wandb=Config.use_wandb,
            )
            avg_return = np.mean(episodic_returns)
            if Config.use_wandb:
                wandb.log(
                    {"charts/val_avg_return": avg_return, "global_step": global_step}
                )
            print(f"Evaluation returns: {episodic_returns}, Average: {avg_return:.2f}")

        if update % Config.save_every == 0 and update > 0:
            model_path = f"runs/{run_name}/models/a2c_model_step_{global_step}.pth"
            torch.save(
                {
                    "actor": actor_network.state_dict(),
                    "critic": critic_network.state_dict(),
                },
                model_path,
            )
            print(f"Model saved at update {update} step {global_step} to {model_path}")

    # --- Final Evaluation and Video Saving ---
    if Config.use_wandb:
        print("Capturing final evaluation video...")
        episodic_returns, eval_frames = evaluate(
            actor_network,
            device,
            Config.env_id,
            env=env,
            seed=Config.seed,
            num_eval_eps=Config.num_eval_episodes,
            record=Config.capture_video,
            grid_env=Config.grid_env,
            atari_wrapper=Config.atari_wrapper,
            env_wrapper=env_wrapper,
            use_wandb=Config.use_wandb,
            render_mode="rgb_array" if Config.capture_video else None,
        )

        # Save final video to file if frames were captured
        if eval_frames:
            train_video_path = "videos/final.mp4"
            imageio.mimsave(train_video_path, eval_frames, fps=30)
            print(f"Final training video saved to {train_video_path}")

    envs.close()
    if Config.use_wandb:
        wandb.finish()

    return actor_network


# --- Main Execution ---
if __name__ == "__main__":
    train_a2c()



import os
import random
import time
from typing import Any, Callable, Optional, Union

import gymnasium as gym
import imageio
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from tqdm import tqdm

import wandb


# ===== CONFIGURATION =====
class Config:
    """Configuration class for PPO training."""

    # Experiment settings
    exp_name: str = "PPO-Experiment"  # Experiment name for logging and saving
    seed: int = 42  # Random seed for reproducibility
    env_id: Optional[str] = "CliffWalking-v0"  # Gymnasium environment ID
    total_timesteps: int = 5_000_000  # Total number of timesteps to train

    # PPO & Agent settings
    lr: float = 3e-4  # Learning rate for optimizer
    gamma: float = 0.99  # Discount factor for rewards
    GAE: float = 0.95  # Generalized Advantage Estimation lambda
    n_envs: int = 1  # Number of parallel environments for data collection
    max_steps: int = 128  # Maximum steps per rollout before updating
    num_minibatches: int = 4  # Number of minibatches for PPO updates
    PPO_EPOCHS: int = 4  # Number of PPO epochs per update
    clip_value: float = 0.2  # PPO clipping value for policy ratio
    ENTROPY_COEFF: float = 0.01  # Entropy coefficient for exploration
    VALUE_COEFF: float = 0.5  # Value loss coefficient in total loss
    num_eval_episodes: int = 5  # Number of evaluation episodes
    anneal_lr: bool = False  # Whether to anneal learning rate over time
    max_grad_norm: float = (
        0.0  # Maximum gradient norm for gradient clipping (0.0 to disable)
    )
    value_clip: bool = False  # Whether to apply value function clipping

    # Logging & Saving
    capture_video: bool = True  # Whether to capture evaluation videos
    use_wandb: bool = True  # Whether to use Weights & Biases for logging
    wandb_project: str = "cleanRL"  # W&B project name
    grid_env: bool = False  # Whether the environment uses discrete grid observations (applies OneHot wrapper)
    eval_every: int = 10000  # Frequency of evaluation during training (in updates)
    save_every: int = 10000  # Frequency of saving the model (in updates)
    atari_wrapper: bool = (
        False  # Whether to apply Atari preprocessing and frame stacking
    )
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = (
        None,
    )  # Optional custom environment wrapper
    normalize_obs: bool = False  # Whether to normalize observations
    normalize_reward: bool = False  # Whether to normalize rewards
    log_gradients: bool = True  # Whether to log gradient norms to W&B
    device: str = "cpu"  # Device for training: "auto", "cpu", "cuda", or "cuda:0" etc.
    custom_agent: Optional[nn.Module] = None  # Custom neural network class or instance
    env: Optional[gym.Env] = None  # Optional pre-created environment


# --- Networks ---
def layer_init(
    layer: nn.Module, std: float = np.sqrt(2), bias_const: float = 0.0
) -> nn.Module:
    torch.nn.init.orthogonal_(layer.weight, std)
    torch.nn.init.constant_(layer.bias, bias_const)
    return layer


class ActorNet(nn.Module):
    def __init__(
        self,
        state_space: Union[int, tuple[int, ...]],
        action_space: int,
    ) -> None:
        super().__init__()

        self.fc1 = layer_init(nn.Linear(state_space, 64))
        self.fc2 = layer_init(nn.Linear(64, 64))
        self.fc3 = layer_init(nn.Linear(64, 32))
        self.out = layer_init(nn.Linear(32, action_space))

    def forward(self, x: torch.Tensor) -> torch.distributions.Distribution:
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))

        logits = self.out(x)
        probs = torch.nn.functional.softmax(logits, dim=-1)
        dist = torch.distributions.Categorical(probs)
        return dist

    def get_action(
        self,
        x: torch.Tensor,
        action: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.distributions.Distribution]:
        dist = self.forward(x)
        if action is None:
            action = dist.sample()

        log_prob = dist.log_prob(action)

        return action, log_prob, dist


class CriticNet(nn.Module):
    def __init__(self, state_space: Union[int, tuple[int, ...]]) -> None:
        super().__init__()
        self.fc1 = layer_init(nn.Linear(state_space, 32))
        self.fc2 = layer_init(nn.Linear(32, 32))
        self.fc3 = layer_init(nn.Linear(32, 16))
        self.value = layer_init(nn.Linear(16, 1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))
        return self.value(x)


class OneHotWrapper(gym.ObservationWrapper):
    def __init__(self, env: gym.Env, obs_shape: int = 16) -> None:
        super().__init__(env)
        self.obs_shape = obs_shape
        self.observation_space = gym.spaces.Box(0, 1, (obs_shape,), dtype=np.float32)

    def observation(self, obs: Any) -> np.ndarray:
        one_hot = torch.zeros(self.obs_shape, dtype=torch.float32)
        one_hot[obs] = 1.0
        return one_hot.numpy()


def make_env(
    env_id: str,
    seed: int,
    idx: int,
    render_mode: Optional[str] = None,
    grid_env: bool = False,
    atari_wrapper: bool = False,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    env: Optional[gym.Env] = None,
) -> Callable[[], gym.Env]:
    def thunk():
        if env is not None:
            # Use provided environment
            env_to_use = env

        else:
            # Create new environment
            env_to_use = gym.make(env_id, render_mode=render_mode)

        env_to_use = gym.wrappers.RecordEpisodeStatistics(env_to_use)

        if grid_env:
            env_to_use = OneHotWrapper(
                env_to_use, obs_shape=env_to_use.observation_space.n
            )

        if Config.normalize_reward:
            env_to_use = gym.wrappers.NormalizeReward(env_to_use)

        if Config.normalize_obs:
            env_to_use = gym.wrappers.NormalizeObservation(env_to_use)
        if atari_wrapper:
            env_to_use = gym.wrappers.AtariPreprocessing(
                env_to_use, grayscale_obs=True, scale_obs=True
            )
            env_to_use = gym.wrappers.FrameStackObservation(env_to_use, stack_size=4)
        if env_wrapper:
            env_to_use = env_wrapper(env_to_use)
        env_to_use.action_space.seed(seed + idx)
        return env_to_use

    return thunk


def calculate_param_norm(model: nn.Module) -> float:
    """Calculate the L2 norm of all parameters in a model."""
    total_norm = 0.0
    for p in model.parameters():
        param_norm = p.data.norm(2)
        total_norm += param_norm.item() ** 2
    return total_norm**0.5


def validate_policy_network_dimensions(
    policy_network: nn.Module, obs_dim: Union[int, tuple[int, ...]], action_dim: int
) -> None:
    """
    Validate that the Policy-network's input and output dimensions match the environment.

    Args:
        policy_network: The neural network model (nn.Module)
        obs_dim: Expected observation dimension (int or tuple)
        action_dim: Expected action dimension
    """
    if isinstance(obs_dim, tuple):
        # For Atari-like, check if it has conv layers
        has_conv = any(
            isinstance(module, nn.Conv2d) for module in policy_network.modules()
        )
        if not has_conv:
            print(
                "Warning: Observation is multi-dimensional but network has no Conv2d layers."
            )
    else:
        # Find first Linear layer for input dimension
        first_layer = None
        for module in policy_network.modules():
            if isinstance(module, nn.Linear):
                first_layer = module
                break
        if first_layer is None:
            raise ValueError(
                "Policy-network must have at least one Linear layer for dimension validation."
            )
        if first_layer.in_features != obs_dim:
            raise ValueError(
                f"Policy-network input dimension {first_layer.in_features} does not match observation dimension {obs_dim}."
            )

    # Find last Linear layer for output dimension
    last_layer = None
    for module in reversed(list(policy_network.modules())):
        if isinstance(module, nn.Linear):
            last_layer = module
            break
    if last_layer is None:
        raise ValueError(
            "Policy-network must have at least one Linear layer for dimension validation."
        )
    if last_layer.out_features != action_dim:
        raise ValueError(
            f"Policy-network output dimension {last_layer.out_features} does not match action dimension {action_dim}."
        )


def validate_critic_network_dimensions(
    critic_network: nn.Module, obs_dim: Union[int, tuple[int, ...]]
) -> None:
    """
    Validate that the Critic-network's input dimension matches the environment.

    Args:
        critic_network: The critic neural network model (nn.Module)
        obs_dim: Expected observation dimension (int or tuple)
    """
    if isinstance(obs_dim, tuple):
        # For Atari-like, check if it has conv layers
        has_conv = any(
            isinstance(module, nn.Conv2d) for module in critic_network.modules()
        )
        if not has_conv:
            print(
                "Warning: Observation is multi-dimensional but network has no Conv2d layers."
            )
    else:
        # Find first Linear layer for input dimension
        first_layer = None
        for module in critic_network.modules():
            if isinstance(module, nn.Linear):
                first_layer = module
                break
        if first_layer is None:
            raise ValueError(
                "Critic-network must have at least one Linear layer for dimension validation."
            )
        if first_layer.in_features != obs_dim:
            raise ValueError(
                f"Critic-network input dimension {first_layer.in_features} does not match observation dimension {obs_dim}."
            )

    # Find last Linear layer for output dimension (should be 1 for value)
    last_layer = None
    for module in reversed(list(critic_network.modules())):
        if isinstance(module, nn.Linear):
            last_layer = module
            break
    if last_layer is None:
        raise ValueError(
            "Critic-network must have at least one Linear layer for dimension validation."
        )
    if last_layer.out_features != 1:
        raise ValueError(
            f"Critic-network output dimension {last_layer.out_features} should be 1 for value estimation."
        )


def validate_feature_network_dimensions(
    feature_network: nn.Module, obs_dim: Union[int, tuple[int, ...]], feature_dim: int
) -> None:
    """
    Validate that the Feature-network's input and output dimensions match expectations.

    Args:
        feature_network: The feature neural network model (nn.Module)
        obs_dim: Expected observation dimension (int or tuple)
        feature_dim: Expected feature dimension
    """
    if isinstance(obs_dim, tuple):
        # For Atari-like, check if it has conv layers
        has_conv = any(
            isinstance(module, nn.Conv2d) for module in feature_network.modules()
        )
        if not has_conv:
            print(
                "Warning: Observation is multi-dimensional but network has no Conv2d layers."
            )
    else:
        # Find first Linear layer for input dimension
        first_layer = None
        for module in feature_network.modules():
            if isinstance(module, nn.Linear):
                first_layer = module
                break
        if first_layer is None:
            raise ValueError(
                "Feature-network must have at least one Linear layer for dimension validation."
            )
        if first_layer.in_features != obs_dim:
            raise ValueError(
                f"Feature-network input dimension {first_layer.in_features} does not match observation dimension {obs_dim}."
            )

    # Find last Linear layer for output dimension
    last_layer = None
    for module in reversed(list(feature_network.modules())):
        if isinstance(module, nn.Linear):
            last_layer = module
            break
    if last_layer is None:
        raise ValueError(
            "Feature-network must have at least one Linear layer for dimension validation."
        )
    if last_layer.out_features != feature_dim:
        raise ValueError(
            f"Feature-network output dimension {last_layer.out_features} does not match expected feature dimension {feature_dim}."
        )


def evaluate(
    model: nn.Module,
    device: torch.device,
    env_id: str,
    env: Optional[gym.Env] = None,
    seed: int = 42,
    num_eval_eps: int = 5,
    record: bool = False,
    render_mode: Optional[str] = None,
    grid_env: bool = False,
    atari_wrapper: bool = False,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    use_wandb: bool = False,
) -> tuple[list[float], list[np.ndarray]]:
    # Create evaluation environment
    eval_env = make_env(
        env_id if env is None else "",
        seed,
        idx=0,
        render_mode=render_mode,
        grid_env=grid_env,
        atari_wrapper=atari_wrapper,
        env_wrapper=env_wrapper,
        env=env,
    )()
    model.eval()
    returns = []
    frames = []

    for _ in tqdm(range(num_eval_eps), desc="Evaluating"):
        obs, _ = eval_env.reset()
        done = False
        rewards = 0.0

        while not done:
            if record:
                frame = eval_env.render()
                frames.append(frame)
            with torch.no_grad():
                obs = torch.tensor(obs, device=device, dtype=torch.float32).unsqueeze(0)
                action, _, _ = model.get_action(obs)
                action = action.cpu().numpy()
                if isinstance(eval_env.action_space, gym.spaces.Discrete):
                    action = action.item()
                else:
                    # For continuous action spaces, squeeze the batch dimension

                    action = action.squeeze(0) if action.shape[0] >= 1 else action

            obs, rewards_curr, terminated, truncated, info = eval_env.step(action)
            done = terminated or truncated
            rewards += rewards_curr

        returns.append(rewards)

    # Save video
    if frames and use_wandb:
        video = np.stack(frames)

        video = np.transpose(video, (0, 3, 1, 2))

        wandb.log(
            {
                "videos/eval_policy": wandb.Video(
                    video,
                    fps=30,
                    format="mp4",
                )
            }
        )
        frames = []
    eval_env.close()
    model.train()

    return returns, frames


def train_ppo(
    env_id: Optional[str] = None,
    env: Optional[gym.Env] = Config.env,
    total_timesteps: int = Config.total_timesteps,
    seed: int = Config.seed,
    lr: float = Config.lr,
    gamma: float = Config.gamma,
    GAE: float = Config.GAE,
    n_envs: int = Config.n_envs,
    max_steps: int = Config.max_steps,
    PPO_EPOCHS: int = Config.PPO_EPOCHS,
    clip_value: float = Config.clip_value,
    ENTROPY_COEFF: float = Config.ENTROPY_COEFF,
    VALUE_COEFF: float = Config.VALUE_COEFF,
    capture_video: bool = Config.capture_video,
    use_wandb: bool = Config.use_wandb,
    wandb_project: str = Config.wandb_project,
    exp_name: str = Config.exp_name,
    grid_env: bool = Config.grid_env,
    max_grad_norm: float = Config.max_grad_norm,
    eval_every: int = Config.eval_every,
    save_every: int = Config.save_every,
    atari_wrapper: bool = Config.atari_wrapper,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    actor_class: Any = ActorNet,
    critic_class: Any = CriticNet,
    num_minibatches: int = Config.num_minibatches,
    log_gradients: bool = Config.log_gradients,
    num_eval_episodes: int = Config.num_eval_episodes,
    anneal_lr: bool = Config.anneal_lr,
    normalize_obs: bool = Config.normalize_obs,
    normalize_reward: bool = Config.normalize_reward,
    device: str = Config.device,
    value_clip: bool = False,
) -> nn.Module:
    # Update Config with passed arguments
    Config.env_id = env_id
    Config.total_timesteps = total_timesteps
    Config.seed = seed
    Config.lr = lr
    Config.gamma = gamma
    Config.GAE = GAE
    Config.n_envs = n_envs
    Config.max_steps = max_steps
    Config.PPO_EPOCHS = PPO_EPOCHS
    Config.clip_value = clip_value
    Config.ENTROPY_COEFF = ENTROPY_COEFF
    Config.VALUE_COEFF = VALUE_COEFF
    Config.capture_video = capture_video
    Config.use_wandb = use_wandb
    Config.wandb_project = wandb_project
    Config.exp_name = exp_name
    Config.grid_env = grid_env
    Config.max_grad_norm = max_grad_norm
    Config.eval_every = eval_every
    Config.save_every = save_every
    Config.atari_wrapper = atari_wrapper
    Config.num_minibatches = num_minibatches
    Config.log_gradients = log_gradients
    Config.num_eval_episodes = num_eval_episodes
    Config.anneal_lr = anneal_lr
    Config.normalize_obs = normalize_obs
    Config.normalize_reward = normalize_reward
    Config.value_clip = value_clip
    Config.device = device

    if env is not None and Config.env_id != env_id:
        raise ValueError(
            "Cannot provide both env_id and env. Use env_id for default environments or env for custom environments."
        )

    if Config.capture_video and not Config.use_wandb:
        raise ValueError(
            "Cannot capture video without WandB enabled. Set use_wandb=True to upload videos."
        )

    run_name = f"{Config.env_id}__{Config.exp_name}__{Config.seed}__{int(time.time())}"

    os.makedirs(f"runs/{run_name}/models", exist_ok=True)

    if Config.use_wandb:
        wandb.init(
            project=Config.wandb_project,
            config=vars(Config()),
            name=run_name,
            monitor_gym=True,
        )

    random.seed(Config.seed)
    np.random.seed(Config.seed)
    torch.manual_seed(Config.seed)

    device = torch.device(Config.device)

    if Config.device == "cuda":
        torch.backends.cudnn.deterministic = True
        torch.cuda.manual_seed(Config.seed)
        torch.cuda.manual_seed_all(Config.seed)
        torch.backends.cudnn.benchmark = False

    # Create environments - check for pre-created env first, then default
    if env is not None:
        env_thunks = [
            make_env(
                "",
                Config.seed,
                i,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                env=env,
                render_mode="rgb_array",
            )
            for i in range(Config.n_envs)
        ]
    else:
        # Use default environment creation
        env_thunks = [
            make_env(
                Config.env_id,
                Config.seed,
                i,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                render_mode="rgb_array",
            )
            for i in range(Config.n_envs)
        ]

    envs = gym.vector.SyncVectorEnv(env_thunks)
    if isinstance(envs.single_observation_space, gym.spaces.Discrete):
        obs_space_shape = (envs.single_observation_space.n,)
    else:
        obs_space_shape = envs.single_observation_space.shape[0]

    action_space_n = (
        envs.single_action_space.n
        if isinstance(envs.single_action_space, gym.spaces.Discrete)
        else envs.single_action_space.shape[0]
    )

    action_shape = (
        ()
        if isinstance(envs.single_action_space, gym.spaces.Discrete)
        else (action_space_n,)
    )

    print(f"Observation Space: {obs_space_shape}, Action Space: {action_space_n}")

    # Create actor network
    if isinstance(actor_class, nn.Module):
        # Use custom actor instance
        validate_policy_network_dimensions(actor_class, obs_space_shape, action_space_n)
        actor_network = actor_class.to(device)
    else:
        # Use actor class
        actor_network = actor_class(obs_space_shape, action_space_n).to(device)

    # Create critic network
    if isinstance(critic_class, nn.Module):
        # Use custom critic instance
        validate_critic_network_dimensions(critic_class, obs_space_shape)
        critic_network = critic_class.to(device)
    else:
        # Use critic class
        critic_network = critic_class(obs_space_shape).to(device)

    print("Actor Network Architecture:")
    print(actor_network)
    print("\nCritic Network Architecture:")
    print(critic_network)

    # Compute derived values from passed parameters
    batch_size = Config.n_envs * Config.max_steps
    minibatch_size = batch_size // Config.num_minibatches
    num_updates = Config.total_timesteps // batch_size

    optimizer = optim.Adam(
        list(actor_network.parameters()) + list(critic_network.parameters()),
        lr=Config.lr,
        eps=1e-5,
    )

    # Tensor Storage

    obs_storage = torch.zeros(
        (Config.max_steps, Config.n_envs) + (obs_space_shape,)
    ).to(device)
    actions_storage = torch.zeros((Config.max_steps, Config.n_envs) + action_shape).to(
        device
    )
    logprobs_storage = torch.zeros((Config.max_steps, Config.n_envs)).to(device)
    rewards_storage = torch.zeros((Config.max_steps, Config.n_envs)).to(device)
    dones_storage = torch.zeros((Config.max_steps, Config.n_envs)).to(device)
    values_storage = torch.zeros((Config.max_steps, Config.n_envs)).to(device)

    global_step = 0

    next_obs, _ = envs.reset(seed=Config.seed)
    next_obs = torch.Tensor(next_obs).to(device)
    next_done = torch.zeros(Config.n_envs).to(device)

    start_time = time.time()

    for update in tqdm(range(1, num_updates + 1), desc="Training Updates"):
        # Annealing the rate if instructed to do so.
        if Config.anneal_lr:
            frac = 1.0 - (update - 1.0) / num_updates
            lrnow = frac * Config.lr
            optimizer.param_groups[0]["lr"] = lrnow

        # Rollout Phase
        for step in range(0, Config.max_steps):
            global_step = (update - 1) * batch_size + step * Config.n_envs
            obs_storage[step] = next_obs

            dones_storage[step] = next_done

            with torch.no_grad():
                action, logprob, dist = actor_network.get_action(next_obs)
                value = critic_network(next_obs)

                # Log distribution statistics
                if Config.use_wandb:
                    if hasattr(dist, "probs"):
                        # Discrete
                        probs = dist.probs
                        wandb.log(
                            {
                                "policy/dist_mean": probs.mean().item(),
                                "policy/dist_std": probs.std().item(),
                                "policy/entropy": dist.entropy().mean().item(),
                                "global_step": global_step,
                            }
                        )
                    else:
                        # Continuous
                        wandb.log(
                            {
                                "policy/dist_mean": dist.mean.mean().item(),
                                "policy/dist_std": dist.stddev.mean().item(),
                                "policy/entropy": dist.entropy().mean().item(),
                                "global_step": global_step,
                            }
                        )

            values_storage[step] = value.flatten()
            # print(action_space_n, action.shape, actions_storage[step].shape, action_shape)
            actions_storage[step] = action
            logprobs_storage[step] = (
                logprob.sum(dim=-1) if len(logprob.shape) > 1 else logprob
            )

            # Step the environment
            new_obs, reward, terminated, truncated, info = envs.step(
                action.cpu().numpy()
            )
            done = np.logical_or(terminated, truncated)

            extrinsic_reward = torch.tensor(reward).to(device)
            rewards_storage[step] = extrinsic_reward
            next_obs = torch.Tensor(new_obs).to(device)
            next_done = torch.Tensor(done).to(device)

        # Calculate returns after the rollout is complete
        with torch.no_grad():
            # Get the bootstrapped value from the state after the last step
            bootstrap_value = critic_network(next_obs).squeeze()

            # Initialize tensors for advantages
            advantages = torch.zeros_like(rewards_storage).to(device)

            # Set the initial "next state" value. If an env was done, this is 0, otherwise it's the bootstrap value.
            lastgae = 0.0

            # Loop backwards from the last step to the first
            for t in reversed(range(Config.max_steps)):
                if t == Config.max_steps - 1:
                    nextnonterminal = 1.0 - next_done
                    next_value = bootstrap_value * nextnonterminal
                else:
                    nextnonterminal = 1.0 - dones_storage[t + 1]
                    next_value = values_storage[t + 1] * nextnonterminal

                delta = (
                    rewards_storage[t] + Config.gamma * next_value - values_storage[t]
                )
                advantages[t] = lastgae = (
                    delta + Config.GAE * lastgae * nextnonterminal * Config.gamma
                )

        # Calculate returns
        returns = advantages + values_storage

        # Normalize advantages
        advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

        # Flatten the batch
        b_obs = obs_storage.reshape((-1,) + (obs_space_shape,))
        b_logprobs = logprobs_storage.reshape(-1)
        b_actions = actions_storage.reshape((-1,) + action_shape)
        b_advantages = advantages.reshape(-1)
        b_returns = returns.reshape(-1)
        b_values = values_storage.reshape(-1)
        clipfracs = []

        # PPO Update Phase
        b_inds = np.arange(batch_size)
        for _epoch in range(Config.PPO_EPOCHS):
            np.random.shuffle(b_inds)
            for start in range(0, batch_size, minibatch_size):
                end = start + minibatch_size
                mb_inds = b_inds[start:end]

                # --- PPO Policy and Value Loss ---

                _, new_log_probs, dist = actor_network.get_action(
                    b_obs[mb_inds], b_actions[mb_inds]
                )
                new_log_probs = (
                    new_log_probs.sum(dim=-1)
                    if len(new_log_probs.shape) > 1
                    else new_log_probs
                )
                logratio = new_log_probs - b_logprobs[mb_inds]
                ratio = logratio.exp()

                with torch.no_grad():
                    # calculate approx_kl http://joschu.net/blog/kl-approx.html
                    old_approx_kl = (-logratio).mean()
                    approx_kl = ((ratio - 1) - logratio).mean()
                    clipfracs += [
                        ((ratio - 1.0).abs() > Config.clip_value).float().mean().item()
                    ]
                    if Config.use_wandb:
                        wandb.log({"charts/approx_kl": approx_kl.item()})

                pg_loss1 = b_advantages[mb_inds] * ratio
                pg_loss2 = b_advantages[mb_inds] * torch.clamp(
                    ratio, 1 - Config.clip_value, 1 + Config.clip_value
                )
                policy_loss = -torch.min(pg_loss1, pg_loss2).mean()

                current_values = critic_network(b_obs[mb_inds])

                # Value clipping
                if Config.value_clip:
                    v_loss_unclipped = (current_values - b_returns[mb_inds]) ** 2
                    v_clipped = b_values[mb_inds] + torch.clamp(
                        current_values - b_values[mb_inds],
                        -Config.clip_value,
                        Config.clip_value,
                    )
                    v_loss_clipped = (v_clipped - b_returns[mb_inds]) ** 2
                    v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)

                    critic_loss = Config.VALUE_COEFF * 0.5 * v_loss_max.mean()

                else:
                    critic_loss = (
                        Config.VALUE_COEFF
                        * 0.5
                        * F.mse_loss(current_values.squeeze(), b_returns[mb_inds])
                    )

                critic_loss = F.mse_loss(current_values.squeeze(), b_returns[mb_inds])
                entropy_loss = dist.entropy().mean()

                # Total Loss
                loss = (
                    policy_loss
                    + Config.VALUE_COEFF * critic_loss
                    - Config.ENTROPY_COEFF * entropy_loss
                )

                optimizer.zero_grad()
                loss.backward()

                # Log gradient norm per layer
                if Config.use_wandb and Config.log_gradients:
                    # Calculate gradient norm before clipping
                    total_norm_before = torch.norm(
                        torch.stack(
                            [
                                torch.norm(p.grad.detach(), 2)
                                for p in list(actor_network.parameters())
                                + list(critic_network.parameters())
                                if p.grad is not None
                            ]
                        ),
                        2,
                    )

                    for name, param in actor_network.named_parameters():
                        if param.grad is not None:
                            grad_norm = torch.norm(param.grad.detach(), 2).item()
                            wandb.log(
                                {
                                    f"gradients/actor_layer_{name}": grad_norm,
                                    "global_step": global_step,
                                }
                            )
                    for name, param in critic_network.named_parameters():
                        if param.grad is not None:
                            grad_norm = torch.norm(param.grad.detach(), 2).item()
                            wandb.log(
                                {
                                    f"gradients/critic_layer_{name}": grad_norm,
                                    "global_step": global_step,
                                }
                            )

                    wandb.log(
                        {
                            "gradients/norm_before_clip": total_norm_before.item(),
                            "global_step": global_step,
                        }
                    )

                # Apply gradient clipping
                if Config.max_grad_norm != 0.0:
                    torch.nn.utils.clip_grad_norm_(
                        list(actor_network.parameters())
                        + list(critic_network.parameters()),
                        max_norm=Config.max_grad_norm,
                    )

                optimizer.step()

        # Log episode returns
        if "episode" in info:
            if Config.n_envs > 1:
                for i in range(Config.n_envs):
                    if done[i]:
                        ep_ret = info["episode"]["r"][i]
                        ep_len = info["episode"]["l"][i]

                        if Config.use_wandb:
                            wandb.log(
                                {
                                    "charts/episodic_return": ep_ret,
                                    "charts/episodic_length": ep_len,
                                    "global_step": global_step,
                                }
                            )
            else:
                if done:
                    ep_ret = info["episode"]["r"]
                    ep_len = info["episode"]["l"]
                    if Config.use_wandb:
                        wandb.log(
                            {
                                "charts/episodic_return": ep_ret,
                                "charts/episodic_length": ep_len,
                                "global_step": global_step,
                            }
                        )

        # Log losses and metrics
        if Config.use_wandb and update % 10 == 0:
            wandb.log(
                {
                    "losses/total_loss": loss.item(),
                    "losses/policy_loss": policy_loss.item(),
                    "losses/value_loss": critic_loss.item(),
                    "charts/learning_rate": optimizer.param_groups[0]["lr"],
                    "rewards/rewards_mean": np.mean(rewards_storage.cpu().numpy()),
                    "rewards/rewards_std": np.std(rewards_storage.cpu().numpy()),
                    "advantages/advantages_mean": b_advantages.mean().item(),
                    "advantages/advantages_std": b_advantages.std().item(),
                    "global_step": global_step,
                    "kl/approx_kl": approx_kl.item(),
                    "kl/old_approx_kl": old_approx_kl.item(),
                    "clipfrac/clip_fraction": np.mean(clipfracs),
                }
            )
            print(
                f"Update {update}, Global Step: {global_step}, Policy Loss: {policy_loss.item():.4f}, Value Loss: {critic_loss.item():.4f}, SPS: {int(global_step / (time.time() - start_time))}"
            )

            if Config.use_wandb:
                wandb.log(
                    {
                        "charts/SPS": int(global_step / (time.time() - start_time)),
                        "charts/global_step": global_step,
                    }
                )
        if update % Config.eval_every == 0:
            episodic_returns, _ = evaluate(
                actor_network,
                device,
                Config.env_id,
                env=env,
                seed=Config.seed,
                num_eval_eps=Config.num_eval_episodes,
                record=Config.capture_video,
                render_mode="rgb_array" if Config.capture_video else None,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                use_wandb=Config.use_wandb,
            )
            avg_return = np.mean(episodic_returns)
            if Config.use_wandb:
                wandb.log(
                    {"charts/val_avg_return": avg_return, "global_step": global_step}
                )
            print(f"Evaluation returns: {episodic_returns}, Average: {avg_return:.2f}")

        if update % Config.save_every == 0 and update > 0:
            model_path = f"runs/{run_name}/models/model_step_{global_step}.pth"
            torch.save(
                {
                    "actor": actor_network.state_dict(),
                    "critic": critic_network.state_dict(),
                },
                model_path,
            )
            print(f"Model saved at update {update} step {global_step} to {model_path}")

    # --- Final Evaluation and Video Saving ---
    if Config.use_wandb:
        print("Capturing final evaluation video...")
        episodic_returns, eval_frames = evaluate(
            actor_network,
            device,
            Config.env_id,
            env=env,
            seed=Config.seed,
            num_eval_eps=Config.num_eval_episodes,
            record=Config.capture_video,
            grid_env=Config.grid_env,
            atari_wrapper=Config.atari_wrapper,
            env_wrapper=env_wrapper,
            use_wandb=Config.use_wandb,
            render_mode="rgb_array" if Config.capture_video else None,
        )

        # Save final video to file if frames were captured
        if eval_frames:
            train_video_path = "videos/final.mp4"
            imageio.mimsave(train_video_path, eval_frames, fps=30)
            print(f"Final training video saved to {train_video_path}")

    envs.close()
    if Config.use_wandb:
        wandb.finish()

    return actor_network


def train_ppo_cnn(
    env_id: Optional[str] = None,
    env: Optional[gym.Env] = Config.env,
    total_timesteps: int = Config.total_timesteps,
    seed: int = Config.seed,
    lr: float = Config.lr,
    gamma: float = Config.gamma,
    GAE: float = Config.GAE,
    n_envs: int = Config.n_envs,
    max_steps: int = Config.max_steps,
    PPO_EPOCHS: int = Config.PPO_EPOCHS,
    clip_value: float = Config.clip_value,
    ENTROPY_COEFF: float = Config.ENTROPY_COEFF,
    VALUE_COEFF: float = Config.VALUE_COEFF,
    capture_video: bool = Config.capture_video,
    use_wandb: bool = Config.use_wandb,
    wandb_project: str = Config.wandb_project,
    exp_name: str = Config.exp_name,
    grid_env: bool = Config.grid_env,
    max_grad_norm: float = Config.max_grad_norm,
    eval_every: int = Config.eval_every,
    save_every: int = Config.save_every,
    atari_wrapper: bool = Config.atari_wrapper,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    actor_class: Any = ActorNet,
    critic_class: Any = CriticNet,
    num_minibatches: int = Config.num_minibatches,
    log_gradients: bool = Config.log_gradients,
    num_eval_episodes: int = Config.num_eval_episodes,
    anneal_lr: bool = Config.anneal_lr,
    normalize_obs: bool = Config.normalize_obs,
    normalize_reward: bool = Config.normalize_reward,
    device: str = Config.device,
    value_clip: bool = False,
) -> nn.Module:
    # Update Config with passed arguments
    Config.env_id = env_id
    Config.total_timesteps = total_timesteps
    Config.seed = seed
    Config.lr = lr
    Config.gamma = gamma
    Config.GAE = GAE
    Config.n_envs = n_envs
    Config.max_steps = max_steps
    Config.PPO_EPOCHS = PPO_EPOCHS
    Config.clip_value = clip_value
    Config.ENTROPY_COEFF = ENTROPY_COEFF
    Config.VALUE_COEFF = VALUE_COEFF
    Config.capture_video = capture_video
    Config.use_wandb = use_wandb
    Config.wandb_project = wandb_project
    Config.exp_name = exp_name
    Config.grid_env = grid_env
    Config.max_grad_norm = max_grad_norm
    Config.eval_every = eval_every
    Config.save_every = save_every
    Config.atari_wrapper = atari_wrapper
    Config.num_minibatches = num_minibatches
    Config.log_gradients = log_gradients
    Config.num_eval_episodes = num_eval_episodes
    Config.anneal_lr = anneal_lr
    Config.normalize_obs = normalize_obs
    Config.normalize_reward = normalize_reward
    Config.value_clip = value_clip
    Config.device = device

    if env is not None and Config.env_id != env_id:
        raise ValueError(
            "Cannot provide both env_id and env. Use env_id for default environments or env for custom environments."
        )

    if Config.capture_video and not Config.use_wandb:
        raise ValueError(
            "Cannot capture video without WandB enabled. Set use_wandb=True to upload videos."
        )

    run_name = f"{Config.env_id}__{Config.exp_name}__{Config.seed}__{int(time.time())}"

    os.makedirs(f"runs/{run_name}/models", exist_ok=True)

    if Config.use_wandb:
        wandb.init(
            project=Config.wandb_project,
            config=vars(Config()),
            name=run_name,
            monitor_gym=True,
        )

    random.seed(Config.seed)
    np.random.seed(Config.seed)
    torch.manual_seed(Config.seed)

    device = torch.device(Config.device)

    if Config.device == "cuda":
        torch.backends.cudnn.deterministic = True
        torch.cuda.manual_seed(Config.seed)
        torch.cuda.manual_seed_all(Config.seed)
        torch.backends.cudnn.benchmark = False

    # Create environments - check for pre-created env first, then default
    if env is not None:
        env_thunks = [
            make_env(
                "",
                Config.seed,
                i,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                env=env,
                render_mode="rgb_array",
            )
            for i in range(Config.n_envs)
        ]
    else:
        # Use default environment creation
        env_thunks = [
            make_env(
                Config.env_id,
                Config.seed,
                i,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                render_mode="rgb_array",
            )
            for i in range(Config.n_envs)
        ]

    envs = gym.vector.SyncVectorEnv(env_thunks)
    if isinstance(envs.single_observation_space, gym.spaces.Discrete):
        obs_space_shape = (envs.single_observation_space.n,)
    else:
        obs_space_shape = envs.single_observation_space.shape

    action_space_n = (
        envs.single_action_space.n
        if isinstance(envs.single_action_space, gym.spaces.Discrete)
        else envs.single_action_space.shape
    )

    action_shape = (
        ()
        if isinstance(envs.single_action_space, gym.spaces.Discrete)
        else (action_space_n,)
    )

    print(f"Observation Space: {obs_space_shape}, Action Space: {action_space_n}")

    # Create actor network
    if isinstance(actor_class, nn.Module):
        # Use custom actor instance
        validate_policy_network_dimensions(actor_class, obs_space_shape, action_space_n)
        actor_network = actor_class.to(device)
    else:
        # Use actor class
        actor_network = actor_class(obs_space_shape, action_space_n).to(device)

    # Create critic network
    if isinstance(critic_class, nn.Module):
        # Use custom critic instance
        validate_critic_network_dimensions(critic_class, obs_space_shape)
        critic_network = critic_class.to(device)
    else:
        # Use critic class
        critic_network = critic_class(obs_space_shape).to(device)

    print("Actor Network Architecture:")
    print(actor_network)
    print("\nCritic Network Architecture:")
    print(critic_network)

    # Compute derived values from passed parameters
    batch_size = Config.n_envs * Config.max_steps
    minibatch_size = batch_size // Config.num_minibatches
    num_updates = Config.total_timesteps // batch_size

    optimizer = optim.Adam(
        list(actor_network.parameters()) + list(critic_network.parameters()),
        lr=Config.lr,
        eps=1e-5,
    )

    # Tensor Storage

    obs_storage = torch.zeros((Config.max_steps, Config.n_envs) + obs_space_shape).to(
        device
    )
    actions_storage = torch.zeros((Config.max_steps, Config.n_envs) + action_shape).to(
        device
    )
    logprobs_storage = torch.zeros((Config.max_steps, Config.n_envs)).to(device)
    rewards_storage = torch.zeros((Config.max_steps, Config.n_envs)).to(device)
    dones_storage = torch.zeros((Config.max_steps, Config.n_envs)).to(device)
    values_storage = torch.zeros((Config.max_steps, Config.n_envs)).to(device)

    global_step = 0

    next_obs, _ = envs.reset(seed=Config.seed)
    next_obs = torch.Tensor(next_obs).to(device)
    next_done = torch.zeros(Config.n_envs).to(device)

    start_time = time.time()

    for update in tqdm(range(1, num_updates + 1), desc="Training Updates"):
        # Annealing the rate if instructed to do so.
        if Config.anneal_lr:
            frac = 1.0 - (update - 1.0) / num_updates
            lrnow = frac * Config.lr
            optimizer.param_groups[0]["lr"] = lrnow

        # Rollout Phase
        for step in range(0, Config.max_steps):
            global_step = (update - 1) * batch_size + step * Config.n_envs
            obs_storage[step] = next_obs

            dones_storage[step] = next_done

            with torch.no_grad():
                action, logprob, dist = actor_network.get_action(next_obs)
                value = critic_network(next_obs)

                # Log distribution statistics
                if Config.use_wandb:
                    if hasattr(dist, "probs"):
                        # Discrete
                        probs = dist.probs
                        wandb.log(
                            {
                                "policy/dist_mean": probs.mean().item(),
                                "policy/dist_std": probs.std().item(),
                                "policy/entropy": dist.entropy().mean().item(),
                                "global_step": global_step,
                            }
                        )
                    else:
                        # Continuous
                        wandb.log(
                            {
                                "policy/dist_mean": dist.mean.mean().item(),
                                "policy/dist_std": dist.stddev.mean().item(),
                                "policy/entropy": dist.entropy().mean().item(),
                                "global_step": global_step,
                            }
                        )

            values_storage[step] = value.flatten()
            # print(action_space_n, action.shape, actions_storage[step].shape, action_shape)
            actions_storage[step] = action
            logprobs_storage[step] = (
                logprob.sum(dim=-1) if len(logprob.shape) > 1 else logprob
            )

            # Step the environment
            new_obs, reward, terminated, truncated, info = envs.step(
                action.cpu().numpy()
            )
            done = np.logical_or(terminated, truncated)

            extrinsic_reward = torch.tensor(reward).to(device)
            rewards_storage[step] = extrinsic_reward
            next_obs = torch.Tensor(new_obs).to(device)
            next_done = torch.Tensor(done).to(device)

        # Calculate returns after the rollout is complete
        with torch.no_grad():
            # Get the bootstrapped value from the state after the last step
            bootstrap_value = critic_network(next_obs).squeeze()

            # Initialize tensors for advantages
            advantages = torch.zeros_like(rewards_storage).to(device)

            # Set the initial "next state" value. If an env was done, this is 0, otherwise it's the bootstrap value.
            lastgae = 0.0

            # Loop backwards from the last step to the first
            for t in reversed(range(Config.max_steps)):
                if t == Config.max_steps - 1:
                    nextnonterminal = 1.0 - next_done
                    next_value = bootstrap_value * nextnonterminal
                else:
                    nextnonterminal = 1.0 - dones_storage[t + 1]
                    next_value = values_storage[t + 1] * nextnonterminal

                delta = (
                    rewards_storage[t] + Config.gamma * next_value - values_storage[t]
                )
                advantages[t] = lastgae = (
                    delta + Config.GAE * lastgae * nextnonterminal * Config.gamma
                )

        # Calculate returns
        returns = advantages + values_storage

        # Normalize advantages
        advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

        # Flatten the batch
        b_obs = obs_storage.reshape((-1,) + obs_space_shape)
        b_logprobs = logprobs_storage.reshape(-1)
        b_actions = actions_storage.reshape((-1,) + action_shape)
        b_advantages = advantages.reshape(-1)
        b_returns = returns.reshape(-1)
        b_values = values_storage.reshape(-1)
        clipfracs = []

        # PPO Update Phase
        b_inds = np.arange(batch_size)
        for _epoch in range(Config.PPO_EPOCHS):
            np.random.shuffle(b_inds)
            for start in range(0, batch_size, minibatch_size):
                end = start + minibatch_size
                mb_inds = b_inds[start:end]

                # --- PPO Policy and Value Loss ---

                _, new_log_probs, dist = actor_network.get_action(
                    b_obs[mb_inds], b_actions[mb_inds]
                )
                logratio = new_log_probs - b_logprobs[mb_inds]
                ratio = logratio.exp()

                with torch.no_grad():
                    # calculate approx_kl http://joschu.net/blog/kl-approx.html
                    old_approx_kl = (-logratio).mean()
                    approx_kl = ((ratio - 1) - logratio).mean()
                    clipfracs += [
                        ((ratio - 1.0).abs() > Config.clip_value).float().mean().item()
                    ]
                    if Config.use_wandb:
                        wandb.log({"charts/approx_kl": approx_kl.item()})

                pg_loss1 = b_advantages[mb_inds] * ratio
                pg_loss2 = b_advantages[mb_inds] * torch.clamp(
                    ratio, 1 - Config.clip_value, 1 + Config.clip_value
                )
                policy_loss = -torch.min(pg_loss1, pg_loss2).mean()

                current_values = critic_network(b_obs[mb_inds])

                # Value clipping
                if Config.value_clip:
                    v_loss_unclipped = (current_values - b_returns[mb_inds]) ** 2
                    v_clipped = b_values[mb_inds] + torch.clamp(
                        current_values - b_values[mb_inds],
                        -Config.clip_value,
                        Config.clip_value,
                    )
                    v_loss_clipped = (v_clipped - b_returns[mb_inds]) ** 2
                    v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)

                    critic_loss = Config.VALUE_COEFF * 0.5 * v_loss_max.mean()

                else:
                    critic_loss = (
                        Config.VALUE_COEFF
                        * 0.5
                        * F.mse_loss(current_values.squeeze(), b_returns[mb_inds])
                    )

                critic_loss = F.mse_loss(current_values.squeeze(), b_returns[mb_inds])
                entropy_loss = dist.entropy().mean()

                # Total Loss
                loss = (
                    policy_loss
                    + Config.VALUE_COEFF * critic_loss
                    - Config.ENTROPY_COEFF * entropy_loss
                )

                optimizer.zero_grad()
                loss.backward()

                # Log gradient norm per layer
                if Config.use_wandb and Config.log_gradients:
                    # Calculate gradient norm before clipping
                    total_norm_before = torch.norm(
                        torch.stack(
                            [
                                torch.norm(p.grad.detach(), 2)
                                for p in list(actor_network.parameters())
                                + list(critic_network.parameters())
                                if p.grad is not None
                            ]
                        ),
                        2,
                    )

                    for name, param in actor_network.named_parameters():
                        if param.grad is not None:
                            grad_norm = torch.norm(param.grad.detach(), 2).item()
                            wandb.log(
                                {
                                    f"gradients/actor_layer_{name}": grad_norm,
                                    "global_step": global_step,
                                }
                            )
                    for name, param in critic_network.named_parameters():
                        if param.grad is not None:
                            grad_norm = torch.norm(param.grad.detach(), 2).item()
                            wandb.log(
                                {
                                    f"gradients/critic_layer_{name}": grad_norm,
                                    "global_step": global_step,
                                }
                            )

                    wandb.log(
                        {
                            "gradients/norm_before_clip": total_norm_before.item(),
                            "global_step": global_step,
                        }
                    )

                # Apply gradient clipping
                if Config.max_grad_norm != 0.0:
                    torch.nn.utils.clip_grad_norm_(
                        list(actor_network.parameters())
                        + list(critic_network.parameters()),
                        max_norm=Config.max_grad_norm,
                    )

                optimizer.step()

        # Log episode returns
        if "episode" in info:
            if Config.n_envs > 1:
                for i in range(Config.n_envs):
                    if done[i]:
                        ep_ret = info["episode"]["r"][i]
                        ep_len = info["episode"]["l"][i]

                        if Config.use_wandb:
                            wandb.log(
                                {
                                    "charts/episodic_return": ep_ret,
                                    "charts/episodic_length": ep_len,
                                    "global_step": global_step,
                                }
                            )
            else:
                if done:
                    ep_ret = info["episode"]["r"]
                    ep_len = info["episode"]["l"]
                    if Config.use_wandb:
                        wandb.log(
                            {
                                "charts/episodic_return": ep_ret,
                                "charts/episodic_length": ep_len,
                                "global_step": global_step,
                            }
                        )

        # Log losses and metrics
        if Config.use_wandb and update % 10 == 0:
            wandb.log(
                {
                    "losses/total_loss": loss.item(),
                    "losses/policy_loss": policy_loss.item(),
                    "losses/value_loss": critic_loss.item(),
                    "charts/learning_rate": optimizer.param_groups[0]["lr"],
                    "rewards/rewards_mean": np.mean(rewards_storage.cpu().numpy()),
                    "rewards/rewards_std": np.std(rewards_storage.cpu().numpy()),
                    "advantages/advantages_mean": b_advantages.mean().item(),
                    "advantages/advantages_std": b_advantages.std().item(),
                    "global_step": global_step,
                    "kl/approx_kl": approx_kl.item(),
                    "kl/old_approx_kl": old_approx_kl.item(),
                    "clipfrac/clip_fraction": np.mean(clipfracs),
                }
            )
            print(
                f"Update {update}, Global Step: {global_step}, Policy Loss: {policy_loss.item():.4f}, Value Loss: {critic_loss.item():.4f}, SPS: {int(global_step / (time.time() - start_time))}"
            )

            if Config.use_wandb:
                wandb.log(
                    {
                        "charts/SPS": int(global_step / (time.time() - start_time)),
                        "charts/global_step": global_step,
                    }
                )
        if update % Config.eval_every == 0:
            episodic_returns, _ = evaluate(
                actor_network,
                device,
                Config.env_id,
                env=env,
                seed=Config.seed,
                num_eval_eps=Config.num_eval_episodes,
                record=Config.capture_video,
                render_mode="rgb_array" if Config.capture_video else None,
                grid_env=Config.grid_env,
                atari_wrapper=Config.atari_wrapper,
                env_wrapper=env_wrapper,
                use_wandb=Config.use_wandb,
            )
            avg_return = np.mean(episodic_returns)
            if Config.use_wandb:
                wandb.log(
                    {"charts/val_avg_return": avg_return, "global_step": global_step}
                )
            print(f"Evaluation returns: {episodic_returns}, Average: {avg_return:.2f}")

        if update % Config.save_every == 0 and update > 0:
            model_path = f"runs/{run_name}/models/model_step_{global_step}.pth"
            torch.save(
                {
                    "actor": actor_network.state_dict(),
                    "critic": critic_network.state_dict(),
                },
                model_path,
            )
            print(f"Model saved at update {update} step {global_step} to {model_path}")

    # --- Final Evaluation and Video Saving ---
    if Config.use_wandb:
        print("Capturing final evaluation video...")
        episodic_returns, eval_frames = evaluate(
            actor_network,
            device,
            Config.env_id,
            env=env,
            seed=Config.seed,
            num_eval_eps=Config.num_eval_episodes,
            record=Config.capture_video,
            grid_env=Config.grid_env,
            atari_wrapper=Config.atari_wrapper,
            env_wrapper=env_wrapper,
            use_wandb=Config.use_wandb,
            render_mode="rgb_array" if Config.capture_video else None,
        )

        # Save final video to file if frames were captured
        if eval_frames:
            train_video_path = "videos/final.mp4"
            imageio.mimsave(train_video_path, eval_frames, fps=30)
            print(f"Final training video saved to {train_video_path}")

    envs.close()
    if Config.use_wandb:
        wandb.finish()

    return actor_network


# --- Main Execution ---
if __name__ == "__main__":
    train_ppo()



import os
import random
import time
from typing import Any, Callable, Optional, Union

import gymnasium as gym
import imageio
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from stable_baselines3.common.buffers import ReplayBuffer
from tqdm import tqdm

import wandb


# ===== CONFIGURATION =====
class Config:
    """Configuration class for DDPG training."""

    # Experiment settings
    exp_name: str = "DDPG-Experiment"
    seed: int = 42
    env_id: Optional[str] = "HalfCheetah-v5"

    low: float = -1.0
    noise_clip: float = 0.5
    high: float = 1.0  # Action space limits for BipedalWalker
    # Training parameters
    total_timesteps: int = 1000000
    learning_rate: float = 3e-4
    buffer_size: int = 100000
    gamma: float = 0.99
    tau: float = 0.005  # Soft update parameter for target networks
    target_network_frequency: int = 50  # How often to update target networks
    batch_size: int = 256

    exploration_fraction: float = 0.1
    learning_starts: int = 25000
    train_frequency: int = 2

    # Logging & Saving
    capture_video: bool = True  # Whether to capture evaluation videos
    use_wandb: bool = True  # Whether to use Weights & Biases for logging
    wandb_project: str = "cleanRL"  # W&B project name
    wandb_entity: str = ""  # Your WandB username/team
    eval_every: int = 500  # Frequency of evaluation during training (in steps)
    save_every: int = 10000  # Frequency of saving the model (in steps)
    num_eval_episodes: int = 10  # Number of evaluation episodes
    normalize_reward: bool = False  # Whether to normalize rewards
    normalize_obs: bool = False  # Whether to normalize observations
    atari_wrapper: bool = False  # Whether to use Atari preprocessing and frame stacking
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = (
        None  # Optional custom environment wrapper
    )
    grid_env: bool = False  # Whether it's a grid environment
    n_envs: int = 1  # Number of parallel environments for data collection
    max_grad_norm: float = (
        0.0  # Maximum gradient norm for gradient clipping (0.0 to disable)
    )
    log_gradients: bool = True  # Whether to log gradient norms to W&B
    device: str = "cpu"  # Device for training: "auto", "cpu", "cuda", or "cuda:0" etc.


class ActorNet(nn.Module):
    def __init__(
        self,
        state_space: Union[int, tuple[int, ...]],
        action_space: int,
    ) -> None:
        super().__init__()
        # Handle state_space as tuple or int
        state_dim = state_space[0] if isinstance(state_space, tuple) else state_space
        print(state_dim)
        print(action_space)
        self.fc1 = nn.Linear(state_dim, 256)
        self.fc2 = nn.Linear(256, 256)
        self.out = nn.Linear(256, action_space)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = torch.tanh(
            self.out(
                torch.nn.functional.mish(
                    self.fc2(torch.nn.functional.mish(self.fc1(x)))
                )
            )
        )
        x = x * 1.0  # Scale to action limits
        return x


class QNet(nn.Module):
    def __init__(self, state_space: Union[int, tuple[int, ...]], action_space: int):
        super().__init__()
        # Handle state_space as tuple or int
        state_dim = state_space[0] if isinstance(state_space, tuple) else state_space

        self.fc1 = nn.Linear(state_dim, 256)
        self.fc2 = nn.Linear(action_space, 256)
        self.fc3 = nn.Linear(512, 512)
        self.reduce = nn.Linear(512, 256)  # Output a single Q-value
        self.out = nn.Linear(256, 1)

    def forward(self, state, act):
        st = torch.nn.functional.mish(self.fc1(state))
        action = torch.nn.functional.mish(self.fc2(act))
        temp = torch.cat((st, action), dim=1)  # Concatenate state and action
        x = torch.nn.functional.mish(self.fc3(temp))
        x = torch.nn.functional.mish(self.reduce(x))
        x = self.out(x)
        return x


class ActorNetCNN(nn.Module):
    def __init__(self, obs_shape, action_space):
        super().__init__()
        # Convolutional layers for image processing
        self.conv1 = nn.Conv2d(obs_shape[0], 32, kernel_size=8, stride=4)
        self.conv2 = nn.Conv2d(32, 64, kernel_size=4, stride=2)
        self.conv3 = nn.Conv2d(64, 64, kernel_size=3, stride=1)

        self.fc1 = nn.Linear(64 * 7 * 7, 512)
        self.out = nn.Linear(512, action_space)

    def forward(self, x):
        # x shape: (batch, channels, height, width)
        x = torch.nn.functional.relu(self.conv1(x))
        x = torch.nn.functional.relu(self.conv2(x))
        x = torch.nn.functional.relu(self.conv3(x))
        x = x.view(x.size(0), -1)  # Flatten
        x = torch.nn.functional.relu(self.fc1(x))
        x = torch.tanh(self.out(x))  # Output between -1 and 1
        x = x * 1.0  # Scale to action limits
        return x


class QNetCNN(nn.Module):
    def __init__(self, obs_shape, action_space):
        super().__init__()
        # Convolutional layers for image processing
        self.conv1 = nn.Conv2d(obs_shape[0], 32, kernel_size=8, stride=4)
        self.conv2 = nn.Conv2d(32, 64, kernel_size=4, stride=2)
        self.conv3 = nn.Conv2d(64, 64, kernel_size=3, stride=1)

        # State processing
        self.state_fc = nn.Linear(64 * 7 * 7, 512)

        # Action processing
        self.action_fc = nn.Linear(action_space, 512)

        # Combined processing
        self.combined_fc = nn.Linear(1024, 512)
        self.out = nn.Linear(512, 1)

    def forward(self, state, action):
        # Process state through conv layers
        x = torch.nn.functional.relu(self.conv1(state))
        x = torch.nn.functional.relu(self.conv2(x))
        x = torch.nn.functional.relu(self.conv3(x))
        x = x.view(x.size(0), -1)  # Flatten
        state_features = torch.nn.functional.relu(self.state_fc(x))

        # Process action
        action_features = torch.nn.functional.relu(self.action_fc(action))

        # Combine state and action features
        combined = torch.cat([state_features, action_features], dim=1)
        x = torch.nn.functional.relu(self.combined_fc(combined))
        return self.out(x)


class OneHotWrapper(gym.ObservationWrapper):
    def __init__(self, env: gym.Env, obs_shape: int = 16) -> None:
        super().__init__(env)
        self.obs_shape = obs_shape
        self.observation_space = gym.spaces.Box(0, 1, (obs_shape,), dtype=np.float32)

    def observation(self, obs: Any) -> np.ndarray:
        one_hot = torch.zeros(self.obs_shape, dtype=torch.float32)
        one_hot[obs] = 1.0
        return one_hot.numpy()


def make_env(
    env_id: str,
    seed: int,
    idx: int,
    render_mode: Optional[str] = None,
    grid_env: bool = False,
    atari_wrapper: bool = False,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    env: Optional[gym.Env] = None,
) -> Callable[[], gym.Env]:
    # Validate that only one of env_id or env is provided
    if env_id and env is not None:
        raise ValueError(
            "Cannot provide both env_id and env. Use env_id for Gymnasium environments or env for custom environments."
        )

    def thunk():
        if env is not None:
            # Use provided environment but still apply wrappers
            env_to_use = env
        else:
            # Create new environment
            env_to_use = gym.make(env_id, render_mode=render_mode)

        # Always apply RecordEpisodeStatistics if not already present
        if env is None:
            env_to_use = gym.wrappers.RecordEpisodeStatistics(env_to_use)

        if grid_env:
            env_to_use = OneHotWrapper(
                env_to_use, obs_shape=env_to_use.observation_space.n
            )

        if Config.normalize_reward:
            env_to_use = gym.wrappers.NormalizeReward(env_to_use)

        if Config.normalize_obs and not atari_wrapper:
            env_to_use = gym.wrappers.NormalizeObservation(env_to_use)
        if atari_wrapper:
            env_to_use = gym.wrappers.AtariPreprocessing(
                env_to_use, grayscale_obs=True, scale_obs=True
            )
            env_to_use = gym.wrappers.FrameStackObservation(env_to_use, stack_size=4)
        if env_wrapper:
            env_to_use = env_wrapper(env_to_use)
        env_to_use.action_space.seed(seed + idx)
        return env_to_use

    return thunk


def calculate_param_norm(model: nn.Module) -> float:
    """Calculate the L2 norm of all parameters in a model."""
    total_norm = 0.0
    for p in model.parameters():
        param_norm = p.data.norm(2)
        total_norm += param_norm.item() ** 2
    return total_norm**0.5


def validate_policy_network_dimensions(
    policy_network: nn.Module, obs_dim: Union[int, tuple[int, ...]], action_dim: int
) -> None:
    """
    Validate that the Policy-network's input and output dimensions match the environment.

    Args:
        policy_network: The neural network model (nn.Module)
        obs_dim: Expected observation dimension (int or tuple)
        action_dim: Expected action dimension
    """
    if isinstance(obs_dim, tuple):
        # For Atari-like, check if it has conv layers
        has_conv = any(
            isinstance(module, nn.Conv2d) for module in policy_network.modules()
        )
        if not has_conv:
            print(
                "Warning: Observation is multi-dimensional but network has no Conv2d layers."
            )
    else:
        # Find first Linear layer for input dimension
        first_layer = None
        for module in policy_network.modules():
            if isinstance(module, nn.Linear):
                first_layer = module
                break
        if first_layer is None:
            raise ValueError(
                "Policy-network must have at least one Linear layer for dimension validation."
            )
        if first_layer.in_features != obs_dim:
            raise ValueError(
                f"Policy-network input dimension {first_layer.in_features} does not match observation dimension {obs_dim}."
            )

    # Find last Linear layer for output dimension
    last_layer = None
    for module in reversed(list(policy_network.modules())):
        if isinstance(module, nn.Linear):
            last_layer = module
            break
    if last_layer is None:
        raise ValueError(
            "Policy-network must have at least one Linear layer for dimension validation."
        )
    if last_layer.out_features != action_dim:
        raise ValueError(
            f"Policy-network output dimension {last_layer.out_features} does not match action dimension {action_dim}."
        )


def validate_critic_network_dimensions(
    critic_network: nn.Module, obs_dim: Union[int, tuple[int, ...]]
) -> None:
    """
    Validate that the Critic-network's input dimension matches the environment.

    Args:
        critic_network: The critic neural network model (nn.Module)
        obs_dim: Expected observation dimension (int or tuple)
        action_dim: Expected action dimension
    """
    if isinstance(obs_dim, tuple):
        # For Atari-like, check if it has conv layers
        has_conv = any(
            isinstance(module, nn.Conv2d) for module in critic_network.modules()
        )
        if not has_conv:
            print(
                "Warning: Observation is multi-dimensional but network has no Conv2d layers."
            )
    else:
        # Find first Linear layer for input dimension
        first_layer = None
        for module in critic_network.modules():
            if isinstance(module, nn.Linear):
                first_layer = module
                break
        if first_layer is None:
            raise ValueError(
                "Critic-network must have at least one Linear layer for dimension validation."
            )
        if first_layer.in_features != obs_dim:
            raise ValueError(
                f"Critic-network input dimension {first_layer.in_features} does not match observation dimension {obs_dim}."
            )

    # Find last Linear layer for output dimension (should be 1 for value)
    last_layer = None
    for module in reversed(list(critic_network.modules())):
        if isinstance(module, nn.Linear):
            last_layer = module
            break
    if last_layer is None:
        raise ValueError(
            "Critic-network must have at least one Linear layer for dimension validation."
        )
    if last_layer.out_features != 1:
        raise ValueError(
            f"Critic-network output dimension {last_layer.out_features} should be 1 for value estimation."
        )


def validate_feature_network_dimensions(
    feature_network: nn.Module, obs_dim: Union[int, tuple[int, ...]], feature_dim: int
) -> None:
    """
    Validate that the Feature-network's input and output dimensions match expectations.

    Args:
        feature_network: The feature neural network model (nn.Module)
        obs_dim: Expected observation dimension (int or tuple)
        feature_dim: Expected feature dimension
    """
    if isinstance(obs_dim, tuple):
        # For Atari-like, check if it has conv layers
        has_conv = any(
            isinstance(module, nn.Conv2d) for module in feature_network.modules()
        )
        if not has_conv:
            print(
                "Warning: Observation is multi-dimensional but network has no Conv2d layers."
            )
    else:
        # Find first Linear layer for input dimension
        first_layer = None
        for module in feature_network.modules():
            if isinstance(module, nn.Linear):
                first_layer = module
                break
        if first_layer is None:
            raise ValueError(
                "Feature-network must have at least one Linear layer for dimension validation."
            )
        if first_layer.in_features != obs_dim:
            raise ValueError(
                f"Feature-network input dimension {first_layer.in_features} does not match observation dimension {obs_dim}."
            )

    # Find last Linear layer for output dimension
    last_layer = None
    for module in reversed(list(feature_network.modules())):
        if isinstance(module, nn.Linear):
            last_layer = module
            break
    if last_layer is None:
        raise ValueError(
            "Feature-network must have at least one Linear layer for dimension validation."
        )
    if last_layer.out_features != feature_dim:
        raise ValueError(
            f"Feature-network output dimension {last_layer.out_features} does not match expected feature dimension {feature_dim}."
        )


def evaluate(
    model: nn.Module,
    device: torch.device,
    env_id: str,
    env: Optional[gym.Env] = None,
    seed: int = 42,
    num_eval_eps: int = 5,
    record: bool = False,
    render_mode: Optional[str] = None,
    grid_env: bool = False,
    atari_wrapper: bool = False,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    use_wandb: bool = False,
) -> tuple[list[float], list[np.ndarray]]:
    # Validate that only one of env_id or env is provided
    if env_id and env is not None:
        raise ValueError(
            "Cannot provide both env_id and env. Use env_id for Gymnasium environments or env for custom environments."
        )

    # Create evaluation environment
    eval_env = make_env(
        env_id if env is None else "",
        seed,
        idx=0,
        render_mode=render_mode,
        grid_env=grid_env,
        atari_wrapper=atari_wrapper,
        env_wrapper=env_wrapper,
        env=env,
    )()
    model.eval()
    returns = []
    frames = []

    for _ in tqdm(range(num_eval_eps), desc="Evaluating"):
        obs, _ = eval_env.reset()
        done = False
        rewards = 0.0

        while not done:
            if record:
                frame = eval_env.render()
                frames.append(frame)
            with torch.no_grad():
                obs_tensor = torch.tensor(
                    np.array(obs), device=device, dtype=torch.float32
                ).unsqueeze(0)
                action = model(obs_tensor)
                action = torch.clip(
                    action, Config.low, Config.high
                )  # Use args low and high
                action = action.cpu().numpy()
                if isinstance(eval_env.action_space, gym.spaces.Discrete):
                    action = action.item()
                else:
                    # For continuous action spaces, ensure it's 1D
                    action = action.flatten()

            obs, rewards_curr, terminated, truncated, info = eval_env.step(action)
            done = terminated or truncated
            rewards += rewards_curr

        returns.append(rewards)

    # Save video
    if frames and use_wandb:
        video = np.stack(frames)

        video = np.transpose(video, (0, 3, 1, 2))

        wandb.log(
            {
                "videos/eval_policy": wandb.Video(
                    video,
                    fps=30,
                    format="mp4",
                )
            }
        )
        frames = []
    eval_env.close()
    model.train()

    return returns, frames


def train_ddpg(
    env_id: Optional[str] = None,
    env: Optional[gym.Env] = None,
    total_timesteps: int = Config.total_timesteps,
    seed: int = Config.seed,
    learning_rate: float = Config.learning_rate,
    buffer_size: int = Config.buffer_size,
    gamma: float = Config.gamma,
    tau: float = Config.tau,
    target_network_frequency: int = Config.target_network_frequency,
    batch_size: int = Config.batch_size,
    exploration_fraction: float = Config.exploration_fraction,
    learning_starts: int = Config.learning_starts,
    train_frequency: int = Config.train_frequency,
    capture_video: bool = Config.capture_video,
    use_wandb: bool = Config.use_wandb,
    wandb_project: str = Config.wandb_project,
    wandb_entity: str = Config.wandb_entity,
    exp_name: str = Config.exp_name,
    eval_every: int = Config.eval_every,
    save_every: int = Config.save_every,
    num_eval_episodes: int = Config.num_eval_episodes,
    n_envs: int = Config.n_envs,
    max_grad_norm: float = Config.max_grad_norm,
    log_gradients: bool = Config.log_gradients,
    device: str = Config.device,
    grid_env: bool = False,
    atari_wrapper: bool = False,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    normalize_obs: bool = Config.normalize_obs,
    normalize_reward: bool = Config.normalize_reward,
    actor_class: Any = ActorNet,
    q_network_class: Any = QNet,
    exploration_noise: float = Config.exploration_fraction,
    noise_clip: float = Config.noise_clip,
    low: float = Config.low,
    high: float = Config.high,
) -> nn.Module:
    # Update Config with passed arguments
    Config.env_id = env_id or env.spec.id
    Config.total_timesteps = total_timesteps
    Config.seed = seed
    Config.learning_rate = learning_rate
    Config.buffer_size = buffer_size
    Config.gamma = gamma
    Config.tau = tau
    Config.target_network_frequency = target_network_frequency
    Config.batch_size = batch_size
    Config.exploration_fraction = exploration_fraction
    Config.learning_starts = learning_starts
    Config.train_frequency = train_frequency
    Config.capture_video = capture_video
    Config.use_wandb = use_wandb
    Config.wandb_project = wandb_project
    Config.wandb_entity = wandb_entity
    Config.exp_name = exp_name
    Config.eval_every = eval_every
    Config.save_every = save_every
    Config.num_eval_episodes = num_eval_episodes
    Config.n_envs = n_envs
    Config.max_grad_norm = max_grad_norm
    Config.log_gradients = log_gradients
    Config.device = device
    Config.grid_env = grid_env
    Config.atari_wrapper = atari_wrapper
    Config.env_wrapper = env_wrapper
    Config.normalize_obs = normalize_obs
    Config.normalize_reward = normalize_reward
    Config.exploration_fraction = exploration_noise
    Config.noise_clip = noise_clip
    Config.low = low
    Config.high = high

    # Validate that only one of env_id or env is provided
    if env_id is not None and env is not None:
        raise ValueError(
            "Cannot provide both env_id and env. Use env_id for Gymnasium environments or env for custom environments."
        )

    if Config.capture_video and not Config.use_wandb:
        raise ValueError(
            "Cannot capture video without WandB enabled. Set use_wandb=True to upload videos."
        )

    run_name = f"{Config.env_id}_{Config.seed}__{int(time.time())}"

    os.makedirs(f"runs/{run_name}/models", exist_ok=True)
    os.makedirs(f"videos/{run_name}/train", exist_ok=True)
    os.makedirs(f"videos/{run_name}/eval", exist_ok=True)
    os.makedirs(f"runs/{run_name}", exist_ok=True)

    if Config.use_wandb:
        wandb.init(
            project=Config.wandb_project,
            entity=Config.wandb_entity,
            config=vars(Config()),
            name=run_name,
            monitor_gym=True,
            save_code=True,
        )

    # Set seeds
    random.seed(Config.seed)
    np.random.seed(Config.seed)
    torch.manual_seed(Config.seed)
    device = torch.device(Config.device)

    # Create environment
    if env is not None:
        env_thunk = make_env(
            "",
            Config.seed,
            idx=0,
            render_mode="rgb_array",
            env_wrapper=Config.env_wrapper,
            env=env,
        )
    else:
        env_thunk = make_env(
            Config.env_id,
            Config.seed,
            idx=0,
            render_mode="rgb_array",
            env_wrapper=Config.env_wrapper,
        )

    env = env_thunk()

    # Determine if we're dealing with discrete observation spaces
    is_discrete_obs = isinstance(env.observation_space, gym.spaces.Discrete)
    obs_space = env.observation_space

    # Compute observation dimensions
    if is_discrete_obs:
        obs_shape = obs_space.n
    else:
        obs_shape = obs_space.shape[0]

    # Compute action dimensions
    action_shape = (
        env.action_space.n
        if isinstance(env.action_space, gym.spaces.Discrete)
        else env.action_space.shape[0]
    )

    # Create actor network
    if isinstance(actor_class, nn.Module):
        # Use custom actor instance
        validate_policy_network_dimensions(actor_class, obs_shape, action_shape)
        actor_net = actor_class.to(device)
    else:
        # Use actor class
        actor_net = actor_class(obs_shape, action_shape).to(device)

    # Create critic network
    if isinstance(q_network_class, nn.Module):
        # Use custom critic instance
        validate_critic_network_dimensions(q_network_class, obs_shape, action_shape)
        q_network = q_network_class.to(device)
    else:
        # Use critic class
        q_network = q_network_class(obs_shape, action_shape).to(device)

    # Create target networks
    target_actor_net = actor_class(obs_shape, action_shape).to(device)
    target_q_network = q_network_class(obs_shape, action_shape).to(device)

    target_q_network.load_state_dict(q_network.state_dict())
    target_actor_net.load_state_dict(actor_net.state_dict())

    # Print network architecture
    print("Actor Network Architecture:")
    print(actor_net)
    print("\nCritic Network Architecture:")
    print(q_network)

    # Optimizers
    actor_optim = optim.Adam(actor_net.parameters(), lr=Config.learning_rate)
    q_optim = optim.Adam(q_network.parameters(), lr=Config.learning_rate)

    # Set networks to training mode
    q_network.train()
    actor_net.train()

    # Replay buffer

    replay_buffer = ReplayBuffer(
        Config.buffer_size,
        env.observation_space,
        env.action_space,
        device=device,
        handle_timeout_termination=False,
        n_envs=Config.n_envs,
    )

    obs, _ = env.reset()
    start_time = time.time()

    for step in tqdm(range(Config.total_timesteps)):
        # Get action from actor network with exploration noise
        with torch.no_grad():
            action = actor_net(
                torch.tensor(obs, device=device, dtype=torch.float32).unsqueeze(0)
            )
            action = action + torch.clip(
                torch.randn_like(action) * Config.exploration_fraction,
                -Config.noise_clip,
                Config.noise_clip,
            )
            action = torch.clip(action, Config.low, Config.high)

        action_np = action.cpu().numpy().flatten()

        new_obs, reward, terminated, truncated, info = env.step(action_np)
        done = np.logical_or(terminated, truncated)
        replay_buffer.add(
            obs, new_obs, action_np, np.array(reward), np.array(done), [info]
        )

        # Training step
        if step > Config.learning_starts:
            data = replay_buffer.sample(Config.batch_size)
            with torch.no_grad():
                next_actions = target_actor_net(
                    data.next_observations.to(torch.float32)
                )
                target_max = target_q_network(
                    data.next_observations.to(torch.float32), next_actions
                )
                td_target = data.rewards + Config.gamma * target_max * (1 - data.dones)

            old_val = q_network(
                data.observations.to(torch.float32), data.actions.to(torch.float32)
            )

            q_optim.zero_grad()
            loss = nn.functional.mse_loss(old_val, td_target)
            loss.backward()

            # Log gradient norm per layer for critic
            if Config.use_wandb and Config.log_gradients:
                for name, param in q_network.named_parameters():
                    if param.grad is not None:
                        grad_norm = torch.norm(param.grad.detach(), 2).item()
                        wandb.log(
                            {
                                f"gradients/critic_layer_{name}": grad_norm,
                                "global_step": step,
                            }
                        )

            # Apply gradient clipping for critic
            if Config.max_grad_norm != 0.0:
                torch.nn.utils.clip_grad_norm_(
                    list(q_network.parameters()),
                    max_norm=Config.max_grad_norm,
                )

            q_optim.step()

            if step % Config.train_frequency == 0:
                actor_optim.zero_grad()
                actions = actor_net(data.observations.to(torch.float32))
                policy_loss = -q_network(
                    data.observations.to(torch.float32), actions.to(torch.float32)
                ).mean()
                policy_loss.backward()

                # Log gradient norm per layer for actor
                if Config.use_wandb and Config.log_gradients:
                    for name, param in actor_net.named_parameters():
                        if param.grad is not None:
                            grad_norm = torch.norm(param.grad.detach(), 2).item()
                            wandb.log(
                                {
                                    f"gradients/actor_layer_{name}": grad_norm,
                                    "global_step": step,
                                }
                            )

                # Apply gradient clipping for actor
                if Config.max_grad_norm != 0.0:
                    torch.nn.utils.clip_grad_norm_(
                        list(actor_net.parameters() + q_network.parameters()),
                        max_norm=Config.max_grad_norm,
                    )

                actor_optim.step()

            # Update target networks
            if step % Config.target_network_frequency == 0:
                for q_params, target_params in zip(
                    q_network.parameters(), target_q_network.parameters()
                ):
                    target_params.data.copy_(
                        Config.tau * q_params.data
                        + (1.0 - Config.tau) * target_params.data
                    )

                for actor_params, target_actor_params in zip(
                    actor_net.parameters(), target_actor_net.parameters()
                ):
                    target_actor_params.data.copy_(
                        Config.tau * actor_params.data
                        + (1.0 - Config.tau) * target_actor_params.data
                    )

            # Log episode returns
            if "episode" in info:
                if Config.n_envs > 1:
                    for i in range(Config.n_envs):
                        if done[i]:
                            ep_ret = info["episode"]["r"][i]
                            ep_len = info["episode"]["l"][i]

                            if Config.use_wandb:
                                wandb.log(
                                    {
                                        "charts/episodic_return": ep_ret,
                                        "charts/episodic_length": ep_len,
                                        "global_step": step,
                                    }
                                )
                else:
                    if done:
                        ep_ret = info["episode"]["r"]
                        ep_len = info["episode"]["l"]
                        if Config.use_wandb:
                            wandb.log(
                                {
                                    "charts/episodic_return": ep_ret,
                                    "charts/episodic_length": ep_len,
                                    "global_step": step,
                                }
                            )

            # Log losses and metrics
            if step % 1000 == 0 and step > Config.learning_starts:
                if Config.use_wandb:
                    wandb.log(
                        {
                            "losses/critic_loss": loss.item(),
                            "losses/actor_loss": policy_loss.item()
                            if step % Config.train_frequency == 0
                            else 0.0,
                            "charts/learning_rate": actor_optim.param_groups[0]["lr"],
                            "rewards/rewards_mean": data.rewards.mean().item(),
                            "rewards/rewards_std": data.rewards.std().item(),
                            "global_step": step,
                        }
                    )
                print(
                    f"Step {step}, Critic Loss: {loss.item():.4f}, Actor Loss: {policy_loss.item() if step % Config.train_frequency == 0 else 0.0:.4f}, SPS: {int(step / (time.time() - start_time))}"
                )

            # Evaluation
            if Config.eval_every > 0 and step % Config.eval_every == 0:
                # eval_env_id = "" if env is not None else Config.env_id
                # eval_env = env
                episodic_returns, eval_frames = evaluate(
                    actor_net,
                    device,
                    Config.env_id,
                    env=None,
                    seed=Config.seed,
                    num_eval_eps=Config.num_eval_episodes,
                    record=Config.capture_video,
                    render_mode="rgb_array" if Config.capture_video else None,
                    grid_env=Config.grid_env,
                    atari_wrapper=Config.atari_wrapper,
                    env_wrapper=Config.env_wrapper,
                    use_wandb=Config.use_wandb,
                )
                avg_return = np.mean(episodic_returns)

                if Config.use_wandb:
                    wandb.log(
                        {
                            "val_episodic_returns": episodic_returns,
                            "charts/val_avg_return": avg_return,
                            "charts/val_return_std": np.std(episodic_returns),
                            "val_step": step,
                        }
                    )
                print(
                    f"Evaluation returns: {[float(r) for r in episodic_returns]}, Average: {avg_return:.2f}"
                )

                # Save video if frames were captured
                if eval_frames and Config.use_wandb:
                    video = np.stack(eval_frames)
                    video = np.transpose(video, (0, 3, 1, 2))
                    wandb.log(
                        {
                            "videos/eval_policy": wandb.Video(
                                video,
                                fps=30,
                                format="mp4",
                            )
                        }
                    )

        # Save model
        if Config.save_every > 0 and step % Config.save_every == 0 and step > 0:
            model_path = f"runs/{run_name}/models/model_step_{step}.pth"
            torch.save(
                {
                    "actor": actor_net.state_dict(),
                    "q_network": q_network.state_dict(),
                    "target_actor": target_actor_net.state_dict(),
                    "target_q_network": target_q_network.state_dict(),
                },
                model_path,
            )
            print(f"Model saved at step {step} to {model_path}")

        if done.all():
            obs, _ = env.reset()
        else:
            obs = new_obs

    # Final evaluation and video saving
    if Config.use_wandb:
        print("Capturing final evaluation video...")
        eval_env_id = "" if env is not None else Config.env_id
        eval_env = env
        episodic_returns, eval_frames = evaluate(
            actor_net,
            device,
            eval_env_id,
            env=eval_env,
            record=True,
            render_mode="rgb_array",
            grid_env=Config.grid_env,
            atari_wrapper=Config.atari_wrapper,
            env_wrapper=Config.env_wrapper,
            use_wandb=Config.use_wandb,
        )

        if eval_frames:
            train_video_path = f"videos/final_{Config.env_id}.mp4"
            imageio.mimsave(train_video_path, eval_frames, fps=30, codec="libx264")
            print(f"Final training video saved to {train_video_path}")
            wandb.finish()

    env.close()
    return actor_net


def train_ddpg_cnn(
    env_id: Optional[str] = None,
    env: Optional[gym.Env] = None,
    total_timesteps: int = Config.total_timesteps,
    seed: int = Config.seed,
    learning_rate: float = Config.learning_rate,
    buffer_size: int = Config.buffer_size,
    gamma: float = Config.gamma,
    tau: float = Config.tau,
    target_network_frequency: int = Config.target_network_frequency,
    batch_size: int = Config.batch_size,
    exploration_fraction: float = Config.exploration_fraction,
    learning_starts: int = Config.learning_starts,
    train_frequency: int = Config.train_frequency,
    exploration_noise: float = Config.exploration_fraction,
    noise_clip: float = Config.noise_clip,
    low: float = Config.low,
    high: float = Config.high,
    capture_video: bool = Config.capture_video,
    use_wandb: bool = Config.use_wandb,
    wandb_project: str = Config.wandb_project,
    wandb_entity: str = Config.wandb_entity,
    exp_name: str = Config.exp_name,
    eval_every: int = Config.eval_every,
    save_every: int = Config.save_every,
    num_eval_episodes: int = Config.num_eval_episodes,
    n_envs: int = Config.n_envs,
    max_grad_norm: float = Config.max_grad_norm,
    log_gradients: bool = Config.log_gradients,
    device: str = Config.device,
    grid_env: bool = False,
    atari_wrapper: bool = False,
    env_wrapper: Optional[Callable[[gym.Env], gym.Env]] = None,
    normalize_reward: bool = Config.normalize_reward,
    actor_class: Any = ActorNet,
    q_network_class: Any = QNet,
) -> nn.Module:
    # Update Config with passed arguments
    Config.env_id = env_id or env.spec.id
    Config.total_timesteps = total_timesteps
    Config.seed = seed
    Config.learning_rate = learning_rate
    Config.buffer_size = buffer_size
    Config.gamma = gamma
    Config.tau = tau
    Config.target_network_frequency = target_network_frequency
    Config.batch_size = batch_size
    Config.exploration_fraction = exploration_fraction
    Config.learning_starts = learning_starts
    Config.train_frequency = train_frequency
    Config.capture_video = capture_video
    Config.use_wandb = use_wandb
    Config.wandb_project = wandb_project
    Config.wandb_entity = wandb_entity
    Config.exp_name = exp_name
    Config.eval_every = eval_every
    Config.save_every = save_every
    Config.num_eval_episodes = num_eval_episodes
    Config.n_envs = n_envs
    Config.max_grad_norm = max_grad_norm
    Config.log_gradients = log_gradients
    Config.device = device
    Config.grid_env = grid_env
    Config.atari_wrapper = atari_wrapper
    Config.env_wrapper = env_wrapper
    Config.exploration_fraction = exploration_noise
    Config.noise_clip = noise_clip
    Config.low = low
    Config.high = high
    Config.normalize_obs = False  # Not supported for CNN-based DDPG
    Config.normalize_reward = normalize_reward

    # Validate that only one of env_id or env is provided
    if env_id is not None and env is not None:
        raise ValueError(
            "Cannot provide both env_id and env. Use env_id for Gymnasium environments or env for custom environments."
        )

    if Config.capture_video and not Config.use_wandb:
        raise ValueError(
            "Cannot capture video without WandB enabled. Set use_wandb=True to upload videos."
        )

    run_name = f"{Config.env_id}__{Config.exp_name}__{Config.seed}__{int(time.time())}"

    os.makedirs(f"runs/{run_name}/models", exist_ok=True)
    os.makedirs(f"videos/{run_name}/train", exist_ok=True)
    os.makedirs(f"videos/{run_name}/eval", exist_ok=True)
    os.makedirs(f"runs/{run_name}", exist_ok=True)

    if Config.use_wandb:
        wandb.init(
            project=Config.wandb_project,
            entity=Config.wandb_entity,
            config=vars(Config()),
            name=run_name,
            monitor_gym=True,
            save_code=True,
        )

    # Set seeds
    random.seed(Config.seed)
    np.random.seed(Config.seed)
    torch.manual_seed(Config.seed)
    device = torch.device(Config.device)

    # Create environment
    if env is not None:
        env_thunk = make_env(
            "",
            Config.seed,
            idx=0,
            render_mode="rgb_array",
            env_wrapper=Config.env_wrapper,
            env=env,
        )
    else:
        env_thunk = make_env(
            Config.env_id,
            Config.seed,
            idx=0,
            render_mode="rgb_array",
            env_wrapper=Config.env_wrapper,
        )

    env = env_thunk()

    # Determine if we're dealing with discrete observation spaces
    is_discrete_obs = isinstance(env.observation_space, gym.spaces.Discrete)
    obs_space = env.observation_space

    # Compute observation dimensions
    if is_discrete_obs:
        obs_shape = obs_space.n
    else:
        obs_shape = obs_space.shape

    # Compute action dimensions
    action_shape = (
        env.action_space.n
        if isinstance(env.action_space, gym.spaces.Discrete)
        else env.action_space.shape[0]
    )

    # Create actor network
    if isinstance(actor_class, nn.Module):
        # Use custom actor instance
        validate_policy_network_dimensions(actor_class, obs_shape, action_shape)
        actor_net = actor_class.to(device)
    else:
        # Use actor class
        actor_net = actor_class(obs_shape, action_shape).to(device)

    # Create critic network
    if isinstance(q_network_class, nn.Module):
        # Use custom critic instance
        validate_critic_network_dimensions(q_network_class, obs_shape, action_shape)
        q_network = q_network_class.to(device)
    else:
        # Use critic class
        q_network = q_network_class(obs_shape, action_shape).to(device)

    actor_net = actor_class(obs_shape, action_shape).to(device)
    q_network = q_network_class(obs_shape, action_shape).to(device)

    # Create target networks
    target_actor_net = actor_class(obs_shape, action_shape).to(device)
    target_q_network = q_network_class(obs_shape, action_shape).to(device)

    target_q_network.load_state_dict(q_network.state_dict())
    target_actor_net.load_state_dict(actor_net.state_dict())

    # Print network architecture
    print("Actor Network Architecture:")
    print(actor_net)
    print("\nCritic Network Architecture:")
    print(q_network)

    # Optimizers
    actor_optim = optim.Adam(actor_net.parameters(), lr=Config.learning_rate)
    q_optim = optim.Adam(q_network.parameters(), lr=Config.learning_rate)

    # Set networks to training mode
    q_network.train()
    actor_net.train()

    # Replay buffer

    replay_buffer = ReplayBuffer(
        Config.buffer_size,
        env.observation_space,
        env.action_space,
        device=device,
        handle_timeout_termination=False,
        n_envs=Config.n_envs,
    )

    obs, _ = env.reset()
    start_time = time.time()

    for step in tqdm(range(Config.total_timesteps)):
        # Get action from actor network with exploration noise
        with torch.no_grad():
            action = actor_net(
                torch.tensor(obs, device=device, dtype=torch.float32).unsqueeze(0)
            )
            action = action + torch.clip(
                torch.randn_like(action) * Config.exploration_fraction,
                -Config.noise_clip,
                Config.noise_clip,
            )
            action = torch.clip(action, Config.low, Config.high)
        action_np = action.cpu().numpy().flatten()

        new_obs, reward, terminated, truncated, info = env.step(action_np)
        done = np.logical_or(terminated, truncated)
        replay_buffer.add(
            obs, new_obs, action_np, np.array(reward), np.array(done), [info]
        )

        # Training step
        if step > Config.learning_starts:
            data = replay_buffer.sample(Config.batch_size)
            with torch.no_grad():
                next_actions = target_actor_net(
                    data.next_observations.to(torch.float32)
                )
                target_max = target_q_network(
                    data.next_observations.to(torch.float32), next_actions
                )
                td_target = data.rewards + Config.gamma * target_max * (1 - data.dones)

            old_val = q_network(
                data.observations.to(torch.float32), data.actions.to(torch.float32)
            )

            q_optim.zero_grad()
            loss = nn.functional.mse_loss(old_val, td_target)
            loss.backward()

            # Log gradient norm per layer for critic
            if Config.use_wandb and Config.log_gradients:
                for name, param in q_network.named_parameters():
                    if param.grad is not None:
                        grad_norm = torch.norm(param.grad.detach(), 2).item()
                        wandb.log(
                            {
                                f"gradients/critic_layer_{name}": grad_norm,
                                "global_step": step,
                            }
                        )

            # Apply gradient clipping for critic
            if Config.max_grad_norm != 0.0:
                torch.nn.utils.clip_grad_norm_(
                    list(q_network.parameters()),
                    max_norm=Config.max_grad_norm,
                )

            q_optim.step()

            if step % Config.train_frequency == 0:
                actor_optim.zero_grad()
                actions = actor_net(data.observations.to(torch.float32))
                policy_loss = -q_network(
                    data.observations.to(torch.float32), actions.to(torch.float32)
                ).mean()
                policy_loss.backward()

                # Log gradient norm per layer for actor
                if Config.use_wandb and Config.log_gradients:
                    for name, param in actor_net.named_parameters():
                        if param.grad is not None:
                            grad_norm = torch.norm(param.grad.detach(), 2).item()
                            wandb.log(
                                {
                                    f"gradients/actor_layer_{name}": grad_norm,
                                    "global_step": step,
                                }
                            )

                # Apply gradient clipping for actor
                if Config.max_grad_norm != 0.0:
                    torch.nn.utils.clip_grad_norm_(
                        list(actor_net.parameters() + q_network.parameters()),
                        max_norm=Config.max_grad_norm,
                    )

                actor_optim.step()

            # Update target networks
            if step % Config.target_network_frequency == 0:
                for q_params, target_params in zip(
                    q_network.parameters(), target_q_network.parameters()
                ):
                    target_params.data.copy_(
                        Config.tau * q_params.data
                        + (1.0 - Config.tau) * target_params.data
                    )

                for actor_params, target_actor_params in zip(
                    actor_net.parameters(), target_actor_net.parameters()
                ):
                    target_actor_params.data.copy_(
                        Config.tau * actor_params.data
                        + (1.0 - Config.tau) * target_actor_params.data
                    )

            # Log episode returns
            if "episode" in info:
                if Config.n_envs > 1:
                    for i in range(Config.n_envs):
                        if done[i]:
                            ep_ret = info["episode"]["r"][i]
                            ep_len = info["episode"]["l"][i]

                            if Config.use_wandb:
                                wandb.log(
                                    {
                                        "charts/episodic_return": ep_ret,
                                        "charts/episodic_length": ep_len,
                                        "global_step": step,
                                    }
                                )
                else:
                    if done:
                        ep_ret = info["episode"]["r"]
                        ep_len = info["episode"]["l"]
                        if Config.use_wandb:
                            wandb.log(
                                {
                                    "charts/episodic_return": ep_ret,
                                    "charts/episodic_length": ep_len,
                                    "global_step": step,
                                }
                            )

            # Log losses and metrics
            if step % 100 == 0:
                if Config.use_wandb:
                    wandb.log(
                        {
                            "losses/critic_loss": loss.item(),
                            "losses/actor_loss": policy_loss.item()
                            if step % Config.train_frequency == 0
                            else 0.0,
                            "charts/learning_rate": actor_optim.param_groups[0]["lr"],
                            "rewards/rewards_mean": data.rewards.mean().item(),
                            "rewards/rewards_std": data.rewards.std().item(),
                            "global_step": step,
                        }
                    )
                print(
                    f"Step {step}, Critic Loss: {loss.item():.4f}, Actor Loss: {policy_loss.item() if step % Config.train_frequency == 0 else 0.0:.4f}, SPS: {int(step / (time.time() - start_time))}"
                )

            # Evaluation
            if Config.eval_every > 0 and step % Config.eval_every == 0:
                # eval_env_id = "" if env is not None else Config.env_id
                # eval_env = env
                episodic_returns, eval_frames = evaluate(
                    actor_net,
                    device,
                    Config.env_id,
                    env=None,
                    seed=Config.seed,
                    num_eval_eps=Config.num_eval_episodes,
                    record=Config.capture_video,
                    render_mode="rgb_array" if Config.capture_video else None,
                    grid_env=Config.grid_env,
                    atari_wrapper=Config.atari_wrapper,
                    env_wrapper=Config.env_wrapper,
                    use_wandb=Config.use_wandb,
                )
                avg_return = np.mean(episodic_returns)

                if Config.use_wandb:
                    wandb.log(
                        {
                            "val_episodic_returns": episodic_returns,
                            "charts/val_avg_return": avg_return,
                            "charts/val_return_std": np.std(episodic_returns),
                            "val_step": step,
                        }
                    )
                print(
                    f"Evaluation returns: {[float(r) for r in episodic_returns]}, Average: {avg_return:.2f}"
                )

                # Save video if frames were captured
                if eval_frames and Config.use_wandb:
                    video = np.stack(eval_frames)
                    video = np.transpose(video, (0, 3, 1, 2))
                    wandb.log(
                        {
                            "videos/eval_policy": wandb.Video(
                                video,
                                fps=30,
                                format="mp4",
                            )
                        }
                    )

        # Save model
        if Config.save_every > 0 and step % Config.save_every == 0 and step > 0:
            model_path = f"runs/{run_name}/models/model_step_{step}.pth"
            torch.save(
                {
                    "actor": actor_net.state_dict(),
                    "q_network": q_network.state_dict(),
                    "target_actor": target_actor_net.state_dict(),
                    "target_q_network": target_q_network.state_dict(),
                },
                model_path,
            )
            print(f"Model saved at step {step} to {model_path}")

        if done.all():
            obs, _ = env.reset()
        else:
            obs = new_obs

    # Final evaluation and video saving
    if Config.use_wandb:
        print("Capturing final evaluation video...")
        eval_env_id = "" if env is not None else Config.env_id
        eval_env = env
        episodic_returns, eval_frames = evaluate(
            actor_net,
            device,
            eval_env_id,
            env=eval_env,
            record=True,
            render_mode="rgb_array",
            grid_env=Config.grid_env,
            atari_wrapper=Config.atari_wrapper,
            env_wrapper=Config.env_wrapper,
            use_wandb=Config.use_wandb,
        )

        if eval_frames:
            train_video_path = f"videos/final_{Config.env_id}.mp4"
            imageio.mimsave(train_video_path, eval_frames, fps=30, codec="libx264")
            print(f"Final training video saved to {train_video_path}")
            wandb.finish()

    env.close()
    return actor_net


# --- Main Execution ---
if __name__ == "__main__":
    train_ddpg()
